# (C) 2001-2024 Intel Corporation. All rights reserved.
# Your use of Intel Corporation's design tools, logic functions and other 
# software and tools, and its AMPP partner logic functions, and any output 
# files from any of the foregoing (including device programming or simulation 
# files), and any associated documentation or information are expressly subject 
# to the terms and conditions of the Intel Program License Subscription 
# Agreement, Intel FPGA IP License Agreement, or other applicable 
# license agreement, including, without limitation, that your use is for the 
# sole purpose of programming logic devices manufactured by Intel and sold by 
# Intel or its authorized distributors.  Please refer to the applicable 
# agreement for further details.


######################################################################
# Test Engine IP - Memory AXI Driver kernels
######################################################################
#
# FILE DESCRIPTION
# ----------------
# This file contains:
# - Tutorial kernels demonstrating usage of the software APIs
# - Kernels for AXI traffic patterns used by memory IP (such as EMIF, HBM, MemSS) design examples.
#
# DISCLAIMERS
# ----------------
# Test Engine IP is provided as a beta version. Its functionality and software APIs are subject to
# change in future releases. Any kernels written using the software APIs will need to be updated if
# the APIs change in the future.
#
# The tutorial kernels provided in this file are not guaranteed to work with all example designs.
# They are only provided to showcase the usage of the software APIs.

import math
import random
import re

from fractions import Fraction
from typing import List, Tuple, Union

import pyhydra.ipkits.hydra.driver_mem_axi4 as driver
import pyhydra.ipkits.hydra.driver_csr_axi4l as driver_csr


# Function aliases for convenience
# Commands
write       = driver.write_cmd         # AXI write command
read        = driver.read_cmd          # AXI read command
wait_writes = driver.wait_writes_cmd   # Wait for all outstanding AXI write responses
wait_reads  = driver.wait_reads_cmd    # Wait for all outstanding AXI read responses
sleep       = driver.sleep_cmd         # Put the driver to sleep
loop        = driver.loop_cmd          # Repeat a sequence of commands
parallel    = driver.parallel_cmd      # Execute commands simultaneously
post        = driver.driver_post_cmd   # Issue a 'post' to other driver(s)
wait        = driver.driver_wait_cmd   # Wait for a 'post' from other driver(s)

# Ops, passed as arguments to a command
wr_worker   = driver.write_worker_op
rd_worker   = driver.read_worker_op

addr_gen    = driver.addr_op
echo        = driver.addr_alu_echo_op
incr        = driver.addr_alu_incr_op
rand        = driver.addr_alu_rand_op

data_eq_addr= driver.data_eq_addr_op
data_eq_id  = driver.data_eq_id_op
data_eq_dq  = driver.data_eq_dq_op
dq_echo     = driver.dq_alu_echo_op
dq_rotate   = driver.dq_alu_rotate_op
dq_invert   = driver.dq_alu_invert_op
dq_prbs     = driver.dq_alu_prbs_op

strb_eq_dm  = driver.strb_eq_dm_op
dm_echo     = driver.dm_alu_echo_op
dm_rotate   = driver.dm_alu_rotate_op
dm_invert   = driver.dm_alu_invert_op
dm_prbs     = driver.dm_alu_prbs_op

# CSR driver commands
write_csr = driver_csr.write_cmd
read_csr  = driver_csr.read_cmd

# Some enums
axburst_fixed = 0b00
axburst_incr  = 0b01
axburst_wrap  = 0b10

xresp_okay   = 0b00
xresp_exokay = 0b01
xresp_slverr = 0b10
xresp_decerr = 0b11

# Address fields. This assumes Memory AXI Driver's ADDR_FIELD_*_MASK params are configured accordingly.
full  = 0
row   = 2
col   = 3
ba    = 4
bg    = 5


class MemAxi4DriverPrograms:
   def __init__(self, config, hydra_ip_name, driver_index):
      ip_config   = config['system']['ips'][hydra_ip_name]
      params      = ip_config['parameters']

      num_drivers = int(params['NUM_DRIVERS']['value'])

      driver_type = params[f'DRIVER_{driver_index}_TYPE_ENUM']['value']
      if driver_type != 'DRIVER_TYPE_MEM_AXI4':
         raise Exception(f'Driver {driver_index} is of type {driver_type}, this class only supports DRIVER_TYPE_MEM_AXI4')

      # Address ALU arg width, impacts the address range accessible by a single read/write command
      self.addr_alu_arg_width = int(params[f'DRIVER_{driver_index}_MEM_AXI4_ADDR_ALU_ARG_WIDTH']['value'])

      # Some parameters related to AXI port widths
      self.max_awid   = 2**int(params[f'DRIVER_{driver_index}_MEM_AXI4_AWID_WIDTH']['value']) - 1
      self.max_arid   = 2**int(params[f'DRIVER_{driver_index}_MEM_AXI4_ARID_WIDTH']['value']) - 1
      self.max_awaddr = 2**int(params[f'DRIVER_{driver_index}_MEM_AXI4_AWADDR_WIDTH']['value']) - 1
      self.max_araddr = 2**int(params[f'DRIVER_{driver_index}_MEM_AXI4_ARADDR_WIDTH']['value']) - 1

      # Carefully handle non-power-of-2 data widths.
      # Ex: if data width = 320 bits, then xDATA width = 256 and xUSER width = 64.
      #     AxADDR needs to be aligned to 256-bit boundary, not 320-bit.
      #     AxSIZE corresponds to only 256-bit xDATA, not 320-bit.
      #
      # Gotcha: Memory AXI Driver IP param 'DRIVER_{driver_index}_MEM_AXI4_RDATA_WIDTH' captures
      # total data width, which would be 320 in the previous example.
      total_wdata_width = int(params[f'DRIVER_{driver_index}_MEM_AXI4_WDATA_WIDTH']['value'])
      wdata_width       = 2 ** math.floor(math.log2(total_wdata_width)) 
      wuser_width       = total_wdata_width - wdata_width               

      total_rdata_width = int(params[f'DRIVER_{driver_index}_MEM_AXI4_WDATA_WIDTH']['value'])
      rdata_width       = 2 ** math.floor(math.log2(total_rdata_width)) 
      ruser_width       = total_rdata_width - rdata_width               

      # Min AxADDR increment for xDATA-aligned accesses = num bytes in an xDATA word (since AxADDR is byte-addressable)
      self.min_awaddr_incr = wdata_width // 8
      self.min_araddr_incr = rdata_width // 8

      # Min column field increment for xDATA-aligned accesses = num DQ-widths in an xDATA word (since column is "DQ-width-addressable")
      #self.min_col_incr = wdata_width // dq_width
      self.min_col_incr = int(params[f'DRIVER_{driver_index}_MEM_AXI4_DATA_DQ_RATIO']['value'])

      # AxSIZE encodes the transfer size, i.e. num valid bits in xDATA.
      # The max transfer size uses all bits in xDATA.
      self.max_awsize = self.clog2(self.min_awaddr_incr)
      self.max_arsize = self.clog2(self.min_araddr_incr)

      # AxQOS is 4-bits wide and AxLEN is 8-bits.
      #NOTE: FP EMIF/HBM uses 2-bit AxQOS and 7-bit AxLEN within the atoms.
      self.max_qos = 2**2 - 1
      self.max_len = 2**7 - 1

      ###########################################################
      # Basic tutorials:
      #
      # Test Engine's Memory AXI Driver software library provides APIs to create kernels for the
      # driver. The kernel is a set of software instructions that primarily represents the
      # AXI traffic sequence. The following APIs are provided:
      #    - write()          ; Generate AXI write commands
      #    - read()           ; Generate AXI read commands
      #    - wait_writes()    ; Wait for all outstanding AXI write responses
      #    - wait_reads()     ; Wait for all outstanding AXI read responses
      #    - parallel()       ; Execute commands simultaneously
      #    - loop()           ; Repeat a sequence of commands
      #    - sleep()          ; Idle for some cycles
      #    - post()           ; Issue a 'post' to other driver(s)
      #    - wait()           ; Wait for a 'post' from other driver(s)
      #
      # This section provides sample kernels that illustrate the usage of these APIs, with the
      # exception of read() and write() which are described under the "Advanced tutorials" section.
      # Instead, in this section a helper function called gencmd() is used to simplify the usage
      # of read() and write().
      ###########################################################

      # Basic AXI traffic generation:
      # Write to some addresses, then read from the same addresses and error check the data
      self.tut1_block_rw = [
         # Write to a specific address with specific data
         *self.gencmd("wr", addr=0x1000, data=0x0123456789ABCDEF),
         # Wait for all write responses to return
         wait_writes(),
         # Read from the same address and check if we get back the same data
         *self.gencmd("rd", addr=0x1000, data=0x0123456789ABCDEF),
         # Wait for all read responses to return
         wait_reads(),

         # Idle for 100 cycles
         sleep(100),

         # Similar to above, but write and read to 10 different addresses
         *self.gencmd("wr", iters=10),
         wait_writes(),
         *self.gencmd("rd", iters=10),
         wait_reads(),

         # Idle for 100 cycles
         sleep(100),

         # Similar to above, but write and read to 10 addresses at random using
         # burst access and random data
         *self.gencmd("wr", iters=10, addr="rand", bl=2, data="prbs7"),
         wait_writes(),
         *self.gencmd("rd", iters=10, addr="rand", bl=2, data="prbs7"),
         wait_reads()
      ]

      # An example test vector - Byte-enable testing:
      # Write twice to every address with different data and a random WSTRB
      # on the first access and inverted WSTRB on the second access.
      # We accomplish this using the below addr and strb pattern presets.
      # These byte-enable presets require bl=1.
      self.tut2_byteen_test = [
         *self.gencmd("wr", iters=10, addr="seq_byteen", data="prbs31", strb="byteen_test"),
         wait_writes(),
         *self.gencmd("rd", iters=10, addr="seq_byteen", data="prbs31", strb="byteen_test"),
         wait_reads()
      ]

      # Instruction-resume:
      # Chain separate instructions to produce a single uninterrupted traffic pattern
      self.tut3_instr_resume = [
         # An instruction can resume the addr/data/strb pattern from where the
         # previous instruction ended (rather than restarting the pattern). This
         # requires the following values to be identical between the instructions
         # that are chained together: addr, data, strb, bl, size
         *self.gencmd("wr", iters=4, addr="rand", bl=2, data="prbs7"),
         *self.gencmd("wr", iters=4, addr="rand", bl=2, data="prbs7", resume=True),
         wait_writes(),
         # This single read command will output the same access pattern as the two
         # write commands that were chained together using instruction-resume
         *self.gencmd("rd", iters=8, addr="rand", bl=2, data="prbs7"),
         wait_reads(),

         # Idle for 100 cycles
         sleep(100),

         # Some example use-cases of using instruction resume:

         # Example 1:
         # You can "randomize/toggle" AXI port values while maintaining an uninterupped
         # pattern on addr/data/strb. Here, we are toggling AxPROT.
         *self.gencmd("wr", addr="rand", prot=0),
         loop(3, [
            *self.gencmd("wr", addr="rand", prot=1, resume=True),
            *self.gencmd("wr", addr="rand", prot=0, resume=True),
         ]),
         wait_writes(),
         *self.gencmd("rd", iters=7, addr="rand"),
         wait_reads(),

         # Idle for 100 cycles
         sleep(100),

         # Example 2:
         # You can decompose a long train of writes and reads into smaller chunks.
         # Here, instead of doing:
         # 8 writes, 8 reads
         # we do:
         # 2 writes, 2 reads, 2 writes, 2 reads, 2 writes, 2 reads, 2 writes, 2 reads
         *self.gencmd("wr", iters=2),
         wait_writes(),
         *self.gencmd("rd", iters=2),
         wait_reads(),
         loop(3, [
            *self.gencmd("wr", iters=2, resume=True),
            wait_writes(),
            *self.gencmd("rd", iters=2, resume=True),
            wait_reads()
         ])
      ]

      # Parallel writes and reads with error checking
      self.tut4_parallel_rw = [
         # Write to an initial address space
         *self.gencmd("wr", iters=10),
         wait_writes(),

         # All instructions within parallel() will execute simultaneously
         parallel([
            # Read from previous address space ...
            *self.gencmd("rd", iters=10),
            # ... and in parallel write to the next address space
            *self.gencmd("wr", iters=10, resume=True)
         ]),
         # Wait for all write responses to return since the next
         # batch of reads will access this same address space.
         # No need to wait for read responses since the next
         # batch of writes will access a different address space.
         wait_writes(),

         # Repeat the above in a loop
         loop(3, [
            parallel([
               *self.gencmd("rd", iters=10, resume=True),
               *self.gencmd("wr", iters=10, resume=True)
            ]),
            wait_writes()
         ]),

         # Read from the final address space
         *self.gencmd("rd", iters=10, resume=True),
         # Wait for all read responses to return
         wait_reads()
      ]

      # Multiple AxIDs:
      # Each AxID corresponds to a distinct traffic stream which can be uniquely customized
      self.tut5_multi_id = [
         # AxIDs 0, 1, and 2 are used. Since iters=8, the AxID port wil output 0, 1, 2, 0, 1, 2, 0, 1
         *self.gencmd("wr", iters=8, id=[0, 1, 2]),
         wait_writes(),
         *self.gencmd("rd", iters=8, id=[0, 1, 2]),
         wait_reads(),

         # The ID sequence can be customized, and the traffic pattern for each ID can be uniquely customized (or not).
         # Here, AWIDs 3 and 4 are used, and since iters=8 the AWID port will output 4, 3, 3, 4, 4, 3, 3, 4
         # AWID=3 uses addr="seq" and bl=2
         # AWID=4 uses addr="rand" and bl=2
         *self.gencmd("wr", iters=8, id=[4, 3, 3, 4], addr=["seq", "rand"], bl=2),
         wait_writes(),
         # To read out the previously written data, we can use different IDs.
         # Here, ARIDs 5 and 6 are used, and since iters=8 the ARID port will output 5, 6, 5, 6, 5, 6, 5, 6
         # AWID=5 uses addr="seq" and bl=2
         # AWID=6 uses addr="rand" and bl=2
         # Note that although the ARIDs differ from the AWIDs in terms of their values and sequence, they
         # represent the same traffic pattern. ARID=5 outputs the same traffic pattern as AWID=3, and ARID=6
         # imitates AWID=4.
         *self.gencmd("rd", iters=8, id=[5, 6],       addr=["seq", "rand"], bl=2),
         wait_reads(),

         # Instruction resume can also be used with multiple AxIDs
         *self.gencmd("wr", iters=4, id=[0, 1], bl=[2, 4]),
         *self.gencmd("wr", iters=4, id=[0, 1], bl=[2, 4], resume=True),
         wait_writes(),
         *self.gencmd("rd", iters=8, id=[0, 1], bl=[2, 4]),
         wait_reads()
      ]

      # Custom bandwidth on the AXI command channels and custom backpressure
      # on the AXI response channels
      self.tut6_custom_bw = [
         # 25% bandwidth on AW channel, 50% bandwidth on W channel, 75% bandwidth (backpressure) on B channel
         *self.gencmd("wr", iters=10, addr_bw=0.25, data_bw=0.50, resp_bw=0.75),
         wait_writes(),
         # 25% bandwidth on AR channel, 50% bandwidth (backpressure) on R channel
         *self.gencmd("rd", iters=10, addr_bw=0.25, resp_bw=0.50),
         wait_reads(),

         # NOTE: On the response channels, the custom bandwidth kicks in after a
         # certain delay from when the first response is received. If this delay
         # is undesirable, you can prelude with a dummy read/write command to
         # configure the backpressure as shown below.

         # Dummy read/write commands with a single iteration and error checking
         # disabled to configure backpressure. Inject some delay after receving
         # the responses to ensure that backpressure is active.
         *self.gencmd("wr", resp_bw=0.50, resp=None),
         *self.gencmd("rd", resp_bw=0.75, resp=None, data=None),
         wait_writes(),
         wait_reads(),
         sleep(200),

         # Longer traffic, using the same backpressure that was configured in
         # the previous dummy commands
         *self.gencmd("wr", iters=10, addr_bw=0.75, data_bw=0.50, resp_bw=0.50),
         wait_writes(),
         *self.gencmd("rd", iters=10, addr_bw=0.50, resp_bw=0.75),
         wait_reads()
      ]

      # This kernel demonstrates driver synchronization.
      # In this demo every even + odd pair of drivers synchronize between each other, and
      # finally all the drivers sync up before executing the final instruction.
      all_drivers = [i for i in range(num_drivers) if i != driver_index]
      if driver_index & 1 == 0 and num_drivers > driver_index + 1:
         odd_driver = driver_index + 1

         # Even driver
         self.tut7_driver_sync = [
            # Write 1s
            *self.gencmd("wr", addr=0x1000, data=0x1111111111111111),

            # The post() command accepts a list of driver objects or driver indices, to which a 'post' will be issued.
            # The other drivers consume the 'post' using a wait() command.
            # post() is a non-blocking command. However if any of the destination drivers' backlog of posts is saturated
            # then this post() command will block until the clogged driver consumes a previous post.
            #
            # In this demo, issue a 'post' to the odd_driver.
            post(odd_driver),

            # Write 2s
            *self.gencmd("wr", addr=0x2000, data=0x2222222222222222),

            # The wait() command accepts a list of driver objects or driver indices, from which a 'post' will be consumed.
            # wait() is a blocking command which waits until at least one 'post' is received from each of the destination
            # drivers.
            #
            # In this demo, wait for a 'post' from the odd_driver.
            wait(odd_driver),

            # Write 4s
            *self.gencmd("wr", addr=0x4000, data=0x4444444444444444),

            # Sync up with all drivers
            parallel([
               post(all_drivers),
               wait(all_drivers)
            ]),

            # Write 5s
            *self.gencmd("wr", addr=0x5000, data=0x5555555555555555),

            wait_writes()
         ]

      elif driver_index & 1 == 1 and num_drivers > driver_index - 1:
         even_driver = driver_index - 1

         # Odd driver
         self.tut7_driver_sync = [
            # Wait for a 'post' from the even_driver.
            wait(even_driver),

            # Write 2s
            *self.gencmd("wr", addr=0x2000, data=0x2222222222222222),

            # Write 3s
            *self.gencmd("wr", addr=0x3000, data=0x3333333333333333),

            # Issue a 'post' to the even_driver.
            post(even_driver),

            # Sync up with all drivers
            parallel([
               post(all_drivers),
               wait(all_drivers)
            ]),

            # Write 5s
            *self.gencmd("wr", addr=0x5000, data=0x5555555555555555),

            wait_writes()
         ]

      else:
         # This driver doesn't have an odd/even pair to synchronize with
         self.tut7_driver_sync = [
            # Sync up with all drivers
            parallel([
               post(all_drivers),
               wait(all_drivers)
            ]),

            # Write 5s
            *self.gencmd("wr", addr=0x5000, data=0x5555555555555555),

            wait_writes()
         ]

      ###########################################################
      # Advanced tutorials:
      #
      # This tutorial section demonstrates usage of read() and write(). The previous
      # "Basic tutorials" section demos the other APIs.
      #
      # Disclaimer:
      # read() and write() are highly versatile allowing you to create a large variety
      # of AXI traffic patterns. However this also implies you'll need to carefully
      # construct the patterns for AXI compliance and to avoid AXI transaction types
      # that are unsupported by the AXI responder. Hence it is stronly recommended to use
      # a helper function such as gencmd() instead. Please refrain from using read() and
      # write() directly unless necessary.
      ###########################################################

      # This kernel demonstrates the read() and write() APIs, and the various AXI signals that can be driven (awvalid/awid/awaddr/...)
      self.adv1_basic_rw = [
         # Write command:
         # 'iters' is number of write commands to issue.
         # READY/VALID signals are specified as (X,Y) which indicates duty-cycle, i.e. X cycles high out of Y total cycles.
         # AWID is specified as a list of values that will show up on the AWID port as is.
         # 'Workers' provide unique traffic streams for each AWID.
         write(iters=20, awvalid=(3,4), wvalid=(1,1), bready=(2,3), awid=[0, 1, 2, 3, 4], workers=[
            # Worker for AWID=0
            wr_worker(
               # Specify values for every signal on the AW, W command channels.
               awid=0, awlen=0, awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0, awaddr=0x0000,
               wdata=0xFEEDBAADCAFEF00D, wstrb=0xFFFFFFFF,
               # Specify expected values for every signal on the B response channel.
               # If the incoming response doesn't match the expected response the error will be logged and flagged.
               bresp=xresp_okay
            ),

            # AxADDR, xDATA, xSTRB have on-chip generators that produce complex patterns.
            # The following workers demonstrate their usage.

            wr_worker(
               # AxADDR = Address ALU generator:
               # 'start' is the start value, 'lo'/'hi' are the lower/upper bounds, 'alu' is a list of ALU ops to derive
               # subsequent addresses, 'align' byte-aligns the address and has identical encoding to AxSIZE. Ex: align=0
               # is 1-byte aligned, align=3 is 8-byte aligned. Hence 'align' also equals num LSBs (of AxADDR) that are zeroed.
               awid=1, awlen=0, awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0, awaddr=addr_gen(start=0x1000, lo=0x1000, hi=0x1FFF, align=0, alu=[incr(self.min_awaddr_incr)]),
               # xDATA = DQ ALU generator:
               # Allows specifying a pattern per DQ pin. If insufficient DQ patterns are provided the compiler
               # replicates the pattern, i.e. if only dq0 and dq1 patterns are provided, the compiler sets
               # dq2=dq0, dq3=dq1, dq4=dq0, dq5=dq1, ... If more DQ patterns than pins are provided the
               # excess is discarded.
               wdata=data_eq_dq(
                  dq0_start =0x5A5A5A5A, dq0_alu =[dq_prbs(7)],
                  dq1_start =0xA5A5A5A5, dq1_alu =[dq_prbs(7)],
                  dq2_start =0x5A5A5A5A, dq2_alu =[dq_prbs(7)],
                  dq3_start =0xA5A5A5A5, dq3_alu =[dq_prbs(7)],
                  dq4_start =0x5A5A5A5A, dq4_alu =[dq_prbs(7)],
                  dq5_start =0xA5A5A5A5, dq5_alu =[dq_prbs(7)],
                  dq6_start =0x5A5A5A5A, dq6_alu =[dq_prbs(7)],
                  dq7_start =0xA5A5A5A5, dq7_alu =[dq_prbs(7)],
                  dq8_start =0x5A5A5A5A, dq8_alu =[dq_prbs(7)],
                  dq9_start =0xA5A5A5A5, dq9_alu =[dq_prbs(7)],
                  dq10_start=0x5A5A5A5A, dq10_alu=[dq_prbs(7)],
                  dq11_start=0xA5A5A5A5, dq11_alu=[dq_prbs(7)],
                  dq12_start=0x5A5A5A5A, dq12_alu=[dq_prbs(7)],
                  dq13_start=0xA5A5A5A5, dq13_alu=[dq_prbs(7)],
                  dq14_start=0x5A5A5A5A, dq14_alu=[dq_prbs(7)],
                  dq15_start=0xA5A5A5A5, dq15_alu=[dq_prbs(7)]
               ),
               # xSTRB = DM ALU generator:
               # Similar to DQ generator, here you can specify patterns for DM pins.
               wstrb=strb_eq_dm(
                  dm0_start =0x0000FFFF, dm0_alu =[dm_rotate()],
                  dm1_start =0xFFFF0000, dm1_alu =[dm_rotate()],
                  dm2_start =0x55555555, dm2_alu =[dm_rotate()],
                  dm3_start =0xAAAAAAAA, dm3_alu =[dm_rotate()],
                  dm4_start =0x0000FFFF, dm4_alu =[dm_rotate()],
                  dm5_start =0xFFFF0000, dm5_alu =[dm_rotate()],
                  dm6_start =0x55555555, dm6_alu =[dm_rotate()],
                  dm7_start =0xAAAAAAAA, dm7_alu =[dm_rotate()]
               ),
               bresp=xresp_okay
            ),

            wr_worker(
               awid=2, awlen=0, awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0, awaddr=addr_gen(start=0x2000, lo=0x2000, hi=0x2FFF, align=0, alu=[incr(self.min_awaddr_incr)]),
               # xDATA = AxADDR:
               # Data contains address which is replicated and nibble-aligned.
               # Ex: if   AxADDR[5:0] = 6'h2C
               #     then xDATA[15:0] = {padding, AxADDR, padding, AxADDR} = {2'h0, 6'h2C, 2'h0, 6'h2C} = 16'h2C_2C
               wdata=data_eq_addr(),
               wstrb=0xFFFFFFFF,
               bresp=xresp_okay
            ),

            wr_worker(
               awid=3, awlen=3, awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0, awaddr=0x3000,
               # xDATA = AxID:
               # Data contains ID which is replicated and nibble-aligned.
               # Ex: if     AxID[2:0] = 3'h7
               #     then xDATA[15:0] = {padding, AxID, padding, AxID, padding, AxID, padding, AxID} = {1'h0, 3'h7, 1'h0, 3'h7, 1'h0, 3'h7, 1'h0, 3'h7} = 16'h77_77
               wdata=data_eq_id(),
               wstrb=0xFFFFFFFF,
               bresp=xresp_okay
            ),

            # Expected response can be a "don't-care", which effectively disables error checking for that signal.
            # The following worker demonstrates its usage.

            wr_worker(
               awid=4, awlen=0, awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0, awaddr=0x4000,
               wdata=0xC0010FF1CED0000D, wstrb=0xFFFFFFFF,
               # BRESP is a "don't-care"
               bresp=None
            )
         ]),

         # Wait for *all* outstanding write responses
         wait_writes(),

         # Read command:
         # Similar API as write command.
         # In this example, reads are configured to be identical to writes.
         # While the worker IDs and issue order are different here, each worker reads the locations that were previously written to in the same order.
         read(iters=20, arvalid=(7,8), rready=(4,5), arid=[0x21, 0x55, 3, 3, 0x7D, 0x7D, 0x21, 0x55, 0x38, 0x38], workers=[
            rd_worker(
               # Specify values for every signal on the AR command channel.
               arid=0x21, arlen=0, arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0, araddr=0x0000,
               # Specify expected values for every signal on the R response channel.
               rdata=0xFEEDBAADCAFEF00D, rstrb=0xFFFFFFFF, rresp=xresp_okay
            ),
            rd_worker(
               arid=0x55, arlen=0, arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0, araddr=addr_gen(start=0x1000, lo=0x1000, hi=0x1FFF, align=0, alu=[incr(self.min_araddr_incr)]),
               rdata=data_eq_dq(
                  dq0_start =0x5A5A5A5A, dq0_alu =[dq_prbs(7)],
                  dq1_start =0xA5A5A5A5, dq1_alu =[dq_prbs(7)],
                  dq2_start =0x5A5A5A5A, dq2_alu =[dq_prbs(7)],
                  dq3_start =0xA5A5A5A5, dq3_alu =[dq_prbs(7)],
                  dq4_start =0x5A5A5A5A, dq4_alu =[dq_prbs(7)],
                  dq5_start =0xA5A5A5A5, dq5_alu =[dq_prbs(7)],
                  dq6_start =0x5A5A5A5A, dq6_alu =[dq_prbs(7)],
                  dq7_start =0xA5A5A5A5, dq7_alu =[dq_prbs(7)],
                  dq8_start =0x5A5A5A5A, dq8_alu =[dq_prbs(7)],
                  dq9_start =0xA5A5A5A5, dq9_alu =[dq_prbs(7)],
                  dq10_start=0x5A5A5A5A, dq10_alu=[dq_prbs(7)],
                  dq11_start=0xA5A5A5A5, dq11_alu=[dq_prbs(7)],
                  dq12_start=0x5A5A5A5A, dq12_alu=[dq_prbs(7)],
                  dq13_start=0xA5A5A5A5, dq13_alu=[dq_prbs(7)],
                  dq14_start=0x5A5A5A5A, dq14_alu=[dq_prbs(7)],
                  dq15_start=0xA5A5A5A5, dq15_alu=[dq_prbs(7)]
               ),
               rstrb=strb_eq_dm(
                  dm0_start =0x0000FFFF, dm0_alu =[dm_rotate()],
                  dm1_start =0xFFFF0000, dm1_alu =[dm_rotate()],
                  dm2_start =0x55555555, dm2_alu =[dm_rotate()],
                  dm3_start =0xAAAAAAAA, dm3_alu =[dm_rotate()],
                  dm4_start =0x0000FFFF, dm4_alu =[dm_rotate()],
                  dm5_start =0xFFFF0000, dm5_alu =[dm_rotate()],
                  dm6_start =0x55555555, dm6_alu =[dm_rotate()],
                  dm7_start =0xAAAAAAAA, dm7_alu =[dm_rotate()]
               ),
               rresp=xresp_okay
            ),
            rd_worker(
               arid=0x7D, arlen=0, arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0, araddr=addr_gen(start=0x2000, lo=0x2000, hi=0x2FFF, align=0, alu=[incr(self.min_araddr_incr)]),
               rdata=data_eq_addr(),
               rstrb=0xFFFFFFFF,
               rresp=xresp_okay
            ),
            rd_worker(
               arid=3, arlen=3, arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0, araddr=0x3000,
               rdata=data_eq_id(),
               rstrb=0xFFFFFFFF,
               rresp=xresp_okay
            ),
            rd_worker(
               arid=0x38, arlen=0, arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0, araddr=0x5000,
               # RDATA and RRESP are "don't-cares". RSTRB cannot be a "don't-care".
               rdata=None, rstrb=0xFFFFFFFF, rresp=None
            )
         ]),

         # Wait for *all* outstanding read responses
         wait_reads(),

         # Sleep for 256 clock cycles
         sleep(256)
      ]

      # This kernel demonstrates the usage of address fields.
      # This feature allows updating selective bits (i.e. fields) in the address.
      # Memory AXI Driver needs to be configured during compile-time (IP generation) for the appropriate number of fields and bitmasks for each field.
      # To update a specific field, specify it as part of the ALU op. Ex: incr(4, 8) increments field #4 by +8
      self.adv2_addr_fields = [
         # Increment the entire address.
         write(iters=8, awvalid=(1,1), wvalid=(1,1), bready=(1,1), awid=[0], workers=[
            wr_worker(
               awid=0, awlen=0, awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0, awaddr=addr_gen(start=0, lo=0, hi=self.max_awaddr, align=0, alu=[incr(self.min_awaddr_incr)]),
               wdata=0xFEEDBAADCAFEF00D, wstrb=0xFFFFFFFF,
               bresp=xresp_okay
            )
         ]),

         # Increment the bank address.
         write(iters=8, awvalid=(1,1), wvalid=(1,1), bready=(1,1), awid=[0], workers=[
            wr_worker(
               awid=0, awlen=0, awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0, awaddr=addr_gen(start=0, lo=0, hi=self.max_awaddr, align=0, alu=[incr(ba, 1)]),
               wdata=0xFEEDBAADCAFEF00D, wstrb=0xFFFFFFFF,
               bresp=xresp_okay
            )
         ]),

         # Increment the row.
         write(iters=8, awvalid=(1,1), wvalid=(1,1), bready=(1,1), awid=[0], workers=[
            wr_worker(
               awid=0, awlen=0, awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0, awaddr=addr_gen(start=0, lo=0, hi=self.max_awaddr, align=0, alu=[incr(row, 1)]),
               wdata=0xFEEDBAADCAFEF00D, wstrb=0xFFFFFFFF,
               bresp=xresp_okay
            )
         ]),

         # Increment the column.
         # Column likely needs to be word-aligned to generate a word-aligned AXI address (in FM6 EMIF atleast).
         write(iters=8, awvalid=(1,1), wvalid=(1,1), bready=(1,1), awid=[0], workers=[
            wr_worker(
               awid=0, awlen=0, awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0, awaddr=addr_gen(start=0, lo=0, hi=self.max_awaddr, align=0, alu=[incr(col, self.min_col_incr)]),
               wdata=0xFEEDBAADCAFEF00D, wstrb=0xFFFFFFFF,
               bresp=xresp_okay
            )
         ]),

         wait_writes()
      ]

      # This kernel demonstrates 'instruction resume'.
      # This feature allows subsequent instructions to continue from where the previous instruction ended.
      # More specifically the on-chip generators can resume pattern generation rather than starting over.
      self.adv3_instr_resume = [
         # The first write command.
         write(iters=7, awvalid=(1,1), wvalid=(1,1), bready=(1,1), awid=[0, 1], workers=[
            # Worker for AWID=0
            wr_worker(
               awid=0, awlen=0, awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0, awaddr=0x0000,
               wdata=0xFEEDBAADCAFEF00D, wstrb=0xFFFFFFFF,
               bresp=xresp_okay
            ),
            # Worker for AWID=1
            # This worker will use the on-chip generators.
            # In the subsequent write command we'll instruct the on-chip generators for this worker to resume.
            wr_worker(
               awid=1, awlen=0, awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0, awaddr=addr_gen(start=0x1000, lo=0x1000, hi=0x1FFF, align=self.max_awsize, alu=[rand(8), incr(self.min_awaddr_incr)]),
               wdata=data_eq_dq(
                  dq0_start =0x5A5A5A5A, dq0_alu =[dq_prbs(7), dq_invert()],
                  dq1_start =0xA5A5A5A5, dq1_alu =[dq_prbs(7), dq_invert()],
                  dq2_start =0x5A5A5A5A, dq2_alu =[dq_prbs(7), dq_invert()],
                  dq3_start =0xA5A5A5A5, dq3_alu =[dq_prbs(7), dq_invert()],
                  dq4_start =0x5A5A5A5A, dq4_alu =[dq_prbs(7), dq_invert()],
                  dq5_start =0xA5A5A5A5, dq5_alu =[dq_prbs(7), dq_invert()],
                  dq6_start =0x5A5A5A5A, dq6_alu =[dq_prbs(7), dq_invert()],
                  dq7_start =0xA5A5A5A5, dq7_alu =[dq_prbs(7), dq_invert()],
                  dq8_start =0x5A5A5A5A, dq8_alu =[dq_prbs(7), dq_invert()],
                  dq9_start =0xA5A5A5A5, dq9_alu =[dq_prbs(7), dq_invert()],
                  dq10_start=0x5A5A5A5A, dq10_alu=[dq_prbs(7), dq_invert()],
                  dq11_start=0xA5A5A5A5, dq11_alu=[dq_prbs(7), dq_invert()],
                  dq12_start=0x5A5A5A5A, dq12_alu=[dq_prbs(7), dq_invert()],
                  dq13_start=0xA5A5A5A5, dq13_alu=[dq_prbs(7), dq_invert()],
                  dq14_start=0x5A5A5A5A, dq14_alu=[dq_prbs(7), dq_invert()],
                  dq15_start=0xA5A5A5A5, dq15_alu=[dq_prbs(7), dq_invert()]
               ),
               wstrb=strb_eq_dm(
                  dm0_start =0x0000FFFF, dm0_alu =[dm_invert(), dm_rotate()],
                  dm1_start =0xFFFF0000, dm1_alu =[dm_invert(), dm_rotate()],
                  dm2_start =0x55555555, dm2_alu =[dm_invert(), dm_rotate()],
                  dm3_start =0xAAAAAAAA, dm3_alu =[dm_invert(), dm_rotate()],
                  dm4_start =0x0000FFFF, dm4_alu =[dm_invert(), dm_rotate()],
                  dm5_start =0xFFFF0000, dm5_alu =[dm_invert(), dm_rotate()],
                  dm6_start =0x55555555, dm6_alu =[dm_invert(), dm_rotate()],
                  dm7_start =0xAAAAAAAA, dm7_alu =[dm_invert(), dm_rotate()]
               ),
               bresp=xresp_okay
            )
         ]),

         # The second write command.
         # Here we'll use 'instruction resume' to continue from where the previous command ended.
         write(iters=5, awvalid=(1,1), wvalid=(1,1), bready=(1,1), awid=[1], workers=[
            # We'll instruct worker for AWID=1 to resume from the previous write command.
            # This is indicated by setting certain values to None, as shown below.
            wr_worker(
               # Here both AxADDR and the ALU ops resume from the previous write command for this worker.
               # Ex: if worker for AWID=1 in the previous write command ended at AWADDR=0x1200 and the last ALU op was: alu=[incr(1), incr(2), incr(3), incr(4)]
               #                                                                                                                      ^^^
               #     then the first output from this instruction will be: AWADDR = 0x1200 * [incr(1), incr(2), incr(3), incr(4)] = 0x1200 + 3 = 0x1203
               #                                                                                               ^^^
               awid=1, awlen=0, awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0, awaddr=addr_gen(start=None, lo=0x1000, hi=0x1FFF, align=self.max_awsize, alu=None),
               # Here both dq_start and dq_alu resume from the previous write command for this worker.
               # If instruction resume is used with DQ generators, all DQ generators must resume.
               wdata=data_eq_dq(
                  dq0_start =None, dq0_alu =None,
                  dq1_start =None, dq1_alu =None,
                  dq2_start =None, dq2_alu =None,
                  dq3_start =None, dq3_alu =None,
                  dq4_start =None, dq4_alu =None,
                  dq5_start =None, dq5_alu =None,
                  dq6_start =None, dq6_alu =None,
                  dq7_start =None, dq7_alu =None,
                  dq8_start =None, dq8_alu =None,
                  dq9_start =None, dq9_alu =None,
                  dq10_start=None, dq10_alu=None,
                  dq11_start=None, dq11_alu=None,
                  dq12_start=None, dq12_alu=None,
                  dq13_start=None, dq13_alu=None,
                  dq14_start=None, dq14_alu=None,
                  dq15_start=None, dq15_alu=None
               ),
               # Here both dm_start and dm_alu resume from the previous write command for this worker.
               # If instruction resume is used with DM generators, all DM generators must resume.
               wstrb=strb_eq_dm(
                  dm0_start =None, dm0_alu =None,
                  dm1_start =None, dm1_alu =None,
                  dm2_start =None, dm2_alu =None,
                  dm3_start =None, dm3_alu =None,
                  dm4_start =None, dm4_alu =None,
                  dm5_start =None, dm5_alu =None,
                  dm6_start =None, dm6_alu =None,
                  dm7_start =None, dm7_alu =None
               ),
               bresp=xresp_okay
            )
         ]),

         # Wait for *all* outstanding write responses
         wait_writes(),

         # First let's read what worker AWID=0 wrote.
         read(iters=4, arvalid=(1,1), rready=(1,1), arid=[4], workers=[
            rd_worker(
               arid=4, arlen=0, arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0, araddr=0x0000,
               rdata=0xFEEDBAADCAFEF00D, rstrb=0xFFFFFFFF, rresp=xresp_okay
            )
         ]),

         # Next let's read what worker AWID=1 wrote, except now we'll use a single read command to verify that
         # 'instruction resume' successfully chained the two separate write commands.
         read(iters=8, arvalid=(1,1), rready=(1,1), arid=[4], workers=[
            rd_worker(
               arid=4, arlen=0, arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0, araddr=addr_gen(start=0x1000, lo=0x1000, hi=0x1FFF, align=self.max_arsize, alu=[rand(8), incr(self.min_araddr_incr)]),
               rdata=data_eq_dq(
                  dq0_start =0x5A5A5A5A, dq0_alu =[dq_prbs(7), dq_invert()],
                  dq1_start =0xA5A5A5A5, dq1_alu =[dq_prbs(7), dq_invert()],
                  dq2_start =0x5A5A5A5A, dq2_alu =[dq_prbs(7), dq_invert()],
                  dq3_start =0xA5A5A5A5, dq3_alu =[dq_prbs(7), dq_invert()],
                  dq4_start =0x5A5A5A5A, dq4_alu =[dq_prbs(7), dq_invert()],
                  dq5_start =0xA5A5A5A5, dq5_alu =[dq_prbs(7), dq_invert()],
                  dq6_start =0x5A5A5A5A, dq6_alu =[dq_prbs(7), dq_invert()],
                  dq7_start =0xA5A5A5A5, dq7_alu =[dq_prbs(7), dq_invert()],
                  dq8_start =0x5A5A5A5A, dq8_alu =[dq_prbs(7), dq_invert()],
                  dq9_start =0xA5A5A5A5, dq9_alu =[dq_prbs(7), dq_invert()],
                  dq10_start=0x5A5A5A5A, dq10_alu=[dq_prbs(7), dq_invert()],
                  dq11_start=0xA5A5A5A5, dq11_alu=[dq_prbs(7), dq_invert()],
                  dq12_start=0x5A5A5A5A, dq12_alu=[dq_prbs(7), dq_invert()],
                  dq13_start=0xA5A5A5A5, dq13_alu=[dq_prbs(7), dq_invert()],
                  dq14_start=0x5A5A5A5A, dq14_alu=[dq_prbs(7), dq_invert()],
                  dq15_start=0xA5A5A5A5, dq15_alu=[dq_prbs(7), dq_invert()]
               ),
               rstrb=strb_eq_dm(
                  dm0_start =0x0000FFFF, dm0_alu =[dm_invert(), dm_rotate()],
                  dm1_start =0xFFFF0000, dm1_alu =[dm_invert(), dm_rotate()],
                  dm2_start =0x55555555, dm2_alu =[dm_invert(), dm_rotate()],
                  dm3_start =0xAAAAAAAA, dm3_alu =[dm_invert(), dm_rotate()],
                  dm4_start =0x0000FFFF, dm4_alu =[dm_invert(), dm_rotate()],
                  dm5_start =0xFFFF0000, dm5_alu =[dm_invert(), dm_rotate()],
                  dm6_start =0x55555555, dm6_alu =[dm_invert(), dm_rotate()],
                  dm7_start =0xAAAAAAAA, dm7_alu =[dm_invert(), dm_rotate()]
               ),
               rresp=xresp_okay
            )
         ]),

         # Wait for *all* outstanding read responses
         wait_reads()
      ]

      ###########################################################
      # IO96B EMIF kernels
      ###########################################################

      # TEST EMIF TG2 TRAFFIC PATTERN EMULATION
      #emif_dq_start = [0x5A5A5A5A, 0xC3C3C3C3, 0x2B2B2B2B, 0x96969696]
      emif_ip_config = None
      for ip_name in config['system']['ips']:
         if config['system']['ips'][ip_name]['type'] == "emif_ph2":
            emif_ip_config = config['system']['ips'][ip_name]

      emif_bl = 1
      emif_bl_wrp = 1
      emif_addr_range = None
      if emif_ip_config is not None:
         pattern = re.compile(".*axi4.*")
         mem_capacity_gbits_list = []
         for interface in emif_ip_config['interfaces']:
            if pattern.search(interface): 
               mem_capacity_gbits_list.append(float(emif_ip_config['interfaces'][interface]['assignments']['mem_capacity_gbits']['value']))
               default_mem_technology = emif_ip_config['interfaces'][interface]['assignments']['default_mem_technology']['value']
         
         mem_capacity_gbits = min(mem_capacity_gbits_list)
         mem_capacity_bytes = int((mem_capacity_gbits / 8) * (2**30))
         max_addr = mem_capacity_bytes - 1
         min_addr = 0x0
         emif_addr_range = (min_addr, max_addr)

         if emif_ip_config['parameters']['MEM_TECHNOLOGY_AUTO_BOOL']['value'] == "false":
            protocol = emif_ip_config['parameters']['MEM_TECHNOLOGY']['value']
         elif emif_ip_config['parameters']['MEM_TECHNOLOGY_AUTO_BOOL']['value'] == "true":
            protocol = default_mem_technology

         if protocol == "MEM_TECHNOLOGY_DDR4":
             emif_bl = 1
             emif_bl_wrp = 1
         elif protocol == "MEM_TECHNOLOGY_DDR5":
            emif_bl = 2         
            emif_bl_wrp = 1     
         elif protocol == "MEM_TECHNOLOGY_LPDDR5":
            emif_bl = 2
            emif_bl_wrp = 2     
         elif protocol == "MEM_TECHNOLOGY_LPDDR4":
             emif_bl = 4        
             emif_bl_wrp = 4    

      self.emif_tg_emulation = [
         # Start with 1 single full W+R
         *self.gencmd(cmd='wr', bl=emif_bl, addr_range=emif_addr_range),
         wait_writes(),
         *self.gencmd(cmd='rd', bl=emif_bl, addr_range=emif_addr_range),
         wait_reads(),

         # 1 single partial W+R
         *self.gencmd(cmd='wr', bl=emif_bl_wrp, addr_range=emif_addr_range),
         wait_writes(),
         *self.gencmd(cmd='rd', bl=emif_bl_wrp, addr_range=emif_addr_range),
         wait_reads(),

         # 512 sequential full write/read
         *self.gencmd(cmd='wr', iters=512, bl=emif_bl, data='prbs7', addr_range=emif_addr_range),
         wait_writes(),
         *self.gencmd(cmd='rd', iters=512, bl=emif_bl, data='prbs7', addr_range=emif_addr_range),
         wait_reads(),

         # 512 sequential partial write/read
         *self.gencmd(cmd='wr', iters=512, bl=emif_bl_wrp, data='prbs7', addr_range=emif_addr_range),
         wait_writes(),
         *self.gencmd(cmd='rd', iters=512, bl=emif_bl_wrp, data='prbs7', addr_range=emif_addr_range),
         wait_reads(),

         # 512 random full write/read
         *self.gencmd(cmd='wr', iters=512, bl=emif_bl, addr='rand', data='prbs7', addr_range=emif_addr_range),
         wait_writes(),
         *self.gencmd(cmd='rd', iters=512, bl=emif_bl, addr='rand', data='prbs7', addr_range=emif_addr_range),
         wait_reads()
      ]

      # Shorter EMIF traffic pattern test
      self.emif_tg_emulation_lite = [
         # 1 full write/read, looped 3 times
         loop(3, [
            *self.gencmd(cmd='wr', bl=emif_bl, addr_range=emif_addr_range),
            wait_writes(),
            *self.gencmd(cmd='rd', bl=emif_bl, addr_range=emif_addr_range),
            wait_reads()
         ]),

         # 1 partial write/read, looped 3 times
         loop(3, [
            *self.gencmd(cmd='wr', bl=emif_bl_wrp, addr_range=emif_addr_range),
            wait_writes(),
            *self.gencmd(cmd='rd', bl=emif_bl_wrp, addr_range=emif_addr_range),
            wait_reads()
         ]),

         # 16 sequential full write/read, looped 8 times
         *self.gencmd(cmd='wr', iters=16, bl=emif_bl, addr='seq', data='prbs7', addr_range=emif_addr_range),
         wait_writes(),
         *self.gencmd(cmd='rd', iters=16, bl=emif_bl, addr='seq', data='prbs7', addr_range=emif_addr_range),
         wait_reads(),
         loop(7, [
            *self.gencmd(cmd='wr', iters=16, bl=emif_bl, addr='seq', addr_range=emif_addr_range, resume=True),
            wait_writes(),
            *self.gencmd(cmd='rd', iters=16, bl=emif_bl, addr='seq', addr_range=emif_addr_range, resume=True),
            wait_reads()
         ]),

         # 16 sequential partial write/read, looped 8 times
         *self.gencmd(cmd='wr', iters=16, bl=emif_bl_wrp, addr='seq', addr_range=emif_addr_range, data='prbs7'),
         wait_writes(),
         *self.gencmd(cmd='rd', iters=16, bl=emif_bl_wrp, addr='seq', addr_range=emif_addr_range, data='prbs7'),
         wait_reads(),
         loop(7, [
            *self.gencmd(cmd='wr', iters=16, bl=emif_bl_wrp, addr='seq', addr_range=emif_addr_range, resume=True),
            wait_writes(),
            *self.gencmd(cmd='rd', iters=16, bl=emif_bl_wrp, addr='seq', addr_range=emif_addr_range, resume=True),
            wait_reads()
         ]),

         # 16 random full write/read, looped 8 times
         *self.gencmd(cmd='wr', iters=16, bl=emif_bl, addr='rand', data='prbs7', addr_range=emif_addr_range),
         wait_writes(),
         *self.gencmd(cmd='rd', iters=16, bl=emif_bl, addr='rand', data='prbs7', addr_range=emif_addr_range),
         wait_reads(),
         loop(7, [
            *self.gencmd(cmd='wr', iters=16, bl=emif_bl, addr='rand', addr_range=emif_addr_range, resume=True),
            wait_writes(),
            *self.gencmd(cmd='rd', iters=16, bl=emif_bl, addr='rand', addr_range=emif_addr_range, resume=True),
            wait_reads()
         ])
      ]

      self.emif_tg_emulation_long = [
         # Start with 1 single full W+R
         *self.gencmd(cmd='wr', bl=emif_bl, addr_range=emif_addr_range),
         wait_writes(),
         *self.gencmd(cmd='rd', bl=emif_bl, addr_range=emif_addr_range),
         wait_reads(),

         # 1 single partial W+R
         *self.gencmd(cmd='wr', bl=emif_bl_wrp, addr_range=emif_addr_range),
         wait_writes(),
         *self.gencmd(cmd='rd', bl=emif_bl_wrp, addr_range=emif_addr_range),
         wait_reads(),

         # 4096 sequential full write/read
         *self.gencmd(cmd='wr', iters=4096, bl=emif_bl, data='prbs7', addr_range=emif_addr_range),
         wait_writes(),
         *self.gencmd(cmd='rd', iters=4096, bl=emif_bl, data='prbs7', addr_range=emif_addr_range),
         wait_reads(),

         # 4096 sequential partial write/read
         *self.gencmd(cmd='wr', iters=4096, bl=emif_bl_wrp, data='prbs7', addr_range=emif_addr_range),
         wait_writes(),
         *self.gencmd(cmd='rd', iters=4096, bl=emif_bl_wrp, data='prbs7', addr_range=emif_addr_range),
         wait_reads(),

         # 4096 random full write/read
         *self.gencmd(cmd='wr', iters=4096, bl=emif_bl, addr='rand', data='prbs7', addr_range=emif_addr_range),
         wait_writes(),
         *self.gencmd(cmd='rd', iters=4096, bl=emif_bl, addr='rand', data='prbs7', addr_range=emif_addr_range),
         wait_reads()
      ]

      # Default PVE VV Traffic Patterns
      # for PVE patterns defined to try different AXI BL
      axibl = 8
      Awsize = self.max_awsize
      Arsize = self.max_arsize
      Awaddr_align = Awsize + self.clog2(axibl)
      #print (Awaddr_align)
      Araddr_align = Arsize + self.clog2(axibl)
      #print (Araddr_align)
      self.pve_vv_prog = [
         # PRBS 31
         write(iters=65536, awvalid=(3,4), wvalid=(1,1), bready=(2,3), awid=[0], workers=[
            # Worker for AWID=0
            wr_worker(
               awid=0, awlen=(axibl-1), awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0,
               awaddr=addr_gen(start=0x0, lo=0x0, hi=0xFFFFFFFF, align=0, alu=[incr(axibl*(self.min_awaddr_incr))]),
               wdata=data_eq_dq(
                  dq0_start =0x5A5A5A5A, dq0_alu =[dq_prbs(31)],
                  dq1_start =0xFFFFFFFF, dq1_alu =[dq_prbs(31)],
                  dq2_start =0xAAAAAAAA, dq2_alu =[dq_prbs(31)],
                  dq3_start =0x55555555, dq3_alu =[dq_prbs(31)],
                  dq4_start =0x12347890, dq4_alu =[dq_prbs(31)],
                  dq5_start =0x23458901, dq5_alu =[dq_prbs(31)],
                  dq6_start =0x89071234, dq6_alu =[dq_prbs(31)],
                  dq7_start =0xABCDE123, dq7_alu =[dq_prbs(31)],
                  dq8_start =0xFFFFAAAA, dq8_alu =[dq_prbs(31)],
                  dq9_start =0x90875431, dq9_alu =[dq_prbs(31)],
                  dq10_start=0x89898989, dq10_alu=[dq_prbs(31)],
                  dq11_start=0x90213456, dq11_alu=[dq_prbs(31)],
                  dq12_start=0x5A678901, dq12_alu=[dq_prbs(31)],
                  dq13_start=0xA5A5A5A5, dq13_alu=[dq_prbs(31)],
                  dq14_start=0xBCDEFFFF, dq14_alu=[dq_prbs(31)],
                  dq15_start=0xAAAA5555, dq15_alu=[dq_prbs(31)]
               ),
               wstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               bresp=xresp_okay
            ),
         ]),
         wait_writes(),
         read(iters=65536, arvalid=(7,8), rready=(4,5), arid=[0], workers=[
            rd_worker(
               arid=0, arlen=(axibl-1), arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0,
               araddr=addr_gen(start=0x0, lo=0x0, hi=0xFFFFFFFF, align=0, alu=[incr(axibl*(self.min_araddr_incr))]),
               rdata=data_eq_dq(
                  dq0_start =0x5A5A5A5A, dq0_alu =[dq_prbs(31)],
                  dq1_start =0xFFFFFFFF, dq1_alu =[dq_prbs(31)],
                  dq2_start =0xAAAAAAAA, dq2_alu =[dq_prbs(31)],
                  dq3_start =0x55555555, dq3_alu =[dq_prbs(31)],
                  dq4_start =0x12347890, dq4_alu =[dq_prbs(31)],
                  dq5_start =0x23458901, dq5_alu =[dq_prbs(31)],
                  dq6_start =0x89071234, dq6_alu =[dq_prbs(31)],
                  dq7_start =0xABCDE123, dq7_alu =[dq_prbs(31)],
                  dq8_start =0xFFFFAAAA, dq8_alu =[dq_prbs(31)],
                  dq9_start =0x90875431, dq9_alu =[dq_prbs(31)],
                  dq10_start=0x89898989, dq10_alu=[dq_prbs(31)],
                  dq11_start=0x90213456, dq11_alu=[dq_prbs(31)],
                  dq12_start=0x5A678901, dq12_alu=[dq_prbs(31)],
                  dq13_start=0xA5A5A5A5, dq13_alu=[dq_prbs(31)],
                  dq14_start=0xBCDEFFFF, dq14_alu=[dq_prbs(31)],
                  dq15_start=0xAAAA5555, dq15_alu=[dq_prbs(31)]
               ),
               rstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               rresp=xresp_okay
            ),
         ]),

         wait_reads(),

         # DATA_EQ_ADDR
         write(iters=65536, awvalid=(1,1), wvalid=(1,1), bready=(2,3), awid=[1], workers=[
            # Worker for AWID=1
            wr_worker(
               awid=1, awlen=(axibl-1), awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0,
               awaddr=addr_gen(start=0x1000000, lo=0x1000000, hi=0xFFFFFFFF, align=0, alu=[incr(axibl*(self.min_awaddr_incr))]),
               wdata=data_eq_addr(), wstrb=0xFFFFFFFF,
               bresp=xresp_okay
            )
         ]),
         wait_writes(),
         read(iters=65536, arvalid=(1,1), rready=(4,5), arid=[1], workers=[
            rd_worker(
               arid=1, arlen=(axibl-1), arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0,
               araddr=addr_gen(start=0x1000000, lo=0x1000000, hi=0xFFFFFFFF, align=0, alu=[incr(axibl*(self.min_araddr_incr))]),
               rdata=data_eq_addr(), rstrb=0xFFFFFFFF, rresp=xresp_okay
            )
         ]),
         wait_reads(),

         # CLOCK Pattern
         write(iters=65536, awvalid=(3,4), wvalid=(1,1), bready=(2,3), awid=[2], workers=[
            # Worker for AWID=2
            wr_worker(
               awid=2, awlen=(axibl-1), awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0,
               awaddr=addr_gen(start=0x2000000, lo=0x2000000, hi=0xFFFFFFFF, align=0, alu=[incr(axibl*(self.min_awaddr_incr))]),
               wdata=data_eq_dq(
                  dq0_start =0x55555555, dq0_alu =[dq_echo()],
                  dq1_start =0x55555555, dq1_alu =[dq_echo()],
                  dq2_start =0x55555555, dq2_alu =[dq_echo()],
                  dq3_start =0x55555555, dq3_alu =[dq_echo()],
                  dq4_start =0x55555555, dq4_alu =[dq_echo()],
                  dq5_start =0x55555555, dq5_alu =[dq_echo()],
                  dq6_start =0x55555555, dq6_alu =[dq_echo()],
                  dq7_start =0x55555555, dq7_alu =[dq_echo()],
                  dq8_start =0x55555555, dq8_alu =[dq_echo()],
                  dq9_start =0x55555555, dq9_alu =[dq_echo()],
                  dq10_start=0x55555555, dq10_alu=[dq_echo()],
                  dq11_start=0x55555555, dq11_alu=[dq_echo()],
                  dq12_start=0x55555555, dq12_alu=[dq_echo()],
                  dq13_start=0x55555555, dq13_alu=[dq_echo()],
                  dq14_start=0x55555555, dq14_alu=[dq_echo()],
                  dq15_start=0x55555555, dq15_alu=[dq_echo()]
               ),
               wstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               bresp=xresp_okay
            ),
         ]),
         wait_writes(),
         read(iters=65536, arvalid=(7,8), rready=(4,5), arid=[2], workers=[
            rd_worker(
               arid=2, arlen=(axibl-1), arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0,
               araddr=addr_gen(start=0x2000000, lo=0x2000000, hi=0xFFFFFFFF, align=0, alu=[incr(axibl*(self.min_araddr_incr))]),
               rdata=data_eq_dq(
                  dq0_start =0x55555555, dq0_alu =[dq_echo()],
                  dq1_start =0x55555555, dq1_alu =[dq_echo()],
                  dq2_start =0x55555555, dq2_alu =[dq_echo()],
                  dq3_start =0x55555555, dq3_alu =[dq_echo()],
                  dq4_start =0x55555555, dq4_alu =[dq_echo()],
                  dq5_start =0x55555555, dq5_alu =[dq_echo()],
                  dq6_start =0x55555555, dq6_alu =[dq_echo()],
                  dq7_start =0x55555555, dq7_alu =[dq_echo()],
                  dq8_start =0x55555555, dq8_alu =[dq_echo()],
                  dq9_start =0x55555555, dq9_alu =[dq_echo()],
                  dq10_start=0x55555555, dq10_alu=[dq_echo()],
                  dq11_start=0x55555555, dq11_alu=[dq_echo()],
                  dq12_start=0x55555555, dq12_alu=[dq_echo()],
                  dq13_start=0x55555555, dq13_alu=[dq_echo()],
                  dq14_start=0x55555555, dq14_alu=[dq_echo()],
                  dq15_start=0x55555555, dq15_alu=[dq_echo()]
               ),
               rstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               rresp=xresp_okay
            ),
         ]),
         wait_reads(),

         # Long 1's followed by 0
         write(iters=65536, awvalid=(3,4), wvalid=(1,1), bready=(2,3), awid=[3], workers=[
            # Worker for AWID=3
            wr_worker(
               awid=3, awlen=(axibl-1), awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0,
               awaddr=addr_gen(start=0x3000000, lo=0x3000000, hi=0xFFFFFFFF, align=0, alu=[incr(axibl*(self.min_awaddr_incr))]),
               wdata=data_eq_dq(
                  dq0_start =0xFEFEFEFE, dq0_alu =[dq_echo()],
                  dq1_start =0xFEFEFEFE, dq1_alu =[dq_echo()],
                  dq2_start =0xFEFEFEFE, dq2_alu =[dq_echo()],
                  dq3_start =0xFEFEFEFE, dq3_alu =[dq_echo()],
                  dq4_start =0xFEFEFEFE, dq4_alu =[dq_echo()],
                  dq5_start =0xFEFEFEFE, dq5_alu =[dq_echo()],
                  dq6_start =0xFEFEFEFE, dq6_alu =[dq_echo()],
                  dq7_start =0xFEFEFEFE, dq7_alu =[dq_echo()],
                  dq8_start =0xFEFEFEFE, dq8_alu =[dq_echo()],
                  dq9_start =0xFEFEFEFE, dq9_alu =[dq_echo()],
                  dq10_start=0xFEFEFEFE, dq10_alu=[dq_echo()],
                  dq11_start=0xFEFEFEFE, dq11_alu=[dq_echo()],
                  dq12_start=0xFEFEFEFE, dq12_alu=[dq_echo()],
                  dq13_start=0xFEFEFEFE, dq13_alu=[dq_echo()],
                  dq14_start=0xFEFEFEFE, dq14_alu=[dq_echo()],
                  dq15_start=0xFEFEFEFE, dq15_alu=[dq_echo()]
               ),
               wstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               bresp=xresp_okay
            ),
         ]),
         wait_writes(),
         read(iters=65536, arvalid=(7,8), rready=(4,5), arid=[3], workers=[
            rd_worker(
               arid=3, arlen=(axibl-1), arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0,
               araddr=addr_gen(start=0x3000000, lo=0x3000000, hi=0xFFFFFFFF, align=0, alu=[incr(axibl*(self.min_araddr_incr))]),
               rdata=data_eq_dq(
                  dq0_start =0xFEFEFEFE, dq0_alu =[dq_echo()],
                  dq1_start =0xFEFEFEFE, dq1_alu =[dq_echo()],
                  dq2_start =0xFEFEFEFE, dq2_alu =[dq_echo()],
                  dq3_start =0xFEFEFEFE, dq3_alu =[dq_echo()],
                  dq4_start =0xFEFEFEFE, dq4_alu =[dq_echo()],
                  dq5_start =0xFEFEFEFE, dq5_alu =[dq_echo()],
                  dq6_start =0xFEFEFEFE, dq6_alu =[dq_echo()],
                  dq7_start =0xFEFEFEFE, dq7_alu =[dq_echo()],
                  dq8_start =0xFEFEFEFE, dq8_alu =[dq_echo()],
                  dq9_start =0xFEFEFEFE, dq9_alu =[dq_echo()],
                  dq10_start=0xFEFEFEFE, dq10_alu=[dq_echo()],
                  dq11_start=0xFEFEFEFE, dq11_alu=[dq_echo()],
                  dq12_start=0xFEFEFEFE, dq12_alu=[dq_echo()],
                  dq13_start=0xFEFEFEFE, dq13_alu=[dq_echo()],
                  dq14_start=0xFEFEFEFE, dq14_alu=[dq_echo()],
                  dq15_start=0xFEFEFEFE, dq15_alu=[dq_echo()]
               ),
               rstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               rresp=xresp_okay
            ),
         ]),
         wait_reads(),

         # Long 0's followed by 1
         write(iters=65536, awvalid=(3,4), wvalid=(1,1), bready=(2,3), awid=[4], workers=[
            # Worker for AWID=4
            wr_worker(
               awid=4, awlen=(axibl-1), awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0,
               awaddr=addr_gen(start=0x4000000, lo=0x4000000, hi=0xFFFFFFFF, align=0, alu=[incr(axibl*(self.min_awaddr_incr))]),
               wdata=data_eq_dq(
                  dq0_start =0x00000001, dq0_alu =[dq_echo()],
                  dq1_start =0x00000001, dq1_alu =[dq_echo()],
                  dq2_start =0x00000001, dq2_alu =[dq_echo()],
                  dq3_start =0x00000001, dq3_alu =[dq_echo()],
                  dq4_start =0x00000001, dq4_alu =[dq_echo()],
                  dq5_start =0x00000001, dq5_alu =[dq_echo()],
                  dq6_start =0x00000001, dq6_alu =[dq_echo()],
                  dq7_start =0x00000001, dq7_alu =[dq_echo()],
                  dq8_start =0x00000001, dq8_alu =[dq_echo()],
                  dq9_start =0x00000001, dq9_alu =[dq_echo()],
                  dq10_start=0x00000001, dq10_alu=[dq_echo()],
                  dq11_start=0x00000001, dq11_alu=[dq_echo()],
                  dq12_start=0x00000001, dq12_alu=[dq_echo()],
                  dq13_start=0x00000001, dq13_alu=[dq_echo()],
                  dq14_start=0x00000001, dq14_alu=[dq_echo()],
                  dq15_start=0x00000001, dq15_alu=[dq_echo()]
               ),
               wstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               bresp=xresp_okay
            ),
         ]),
         wait_writes(),
         read(iters=65536, arvalid=(7,8), rready=(4,5), arid=[4], workers=[
            rd_worker(
               arid=4, arlen=(axibl-1), arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0,
               araddr=addr_gen(start=0x4000000, lo=0x4000000, hi=0xFFFFFFFF, align=0, alu=[incr(axibl*(self.min_araddr_incr))]),
               rdata=data_eq_dq(
                  dq0_start =0x00000001, dq0_alu =[dq_echo()],
                  dq1_start =0x00000001, dq1_alu =[dq_echo()],
                  dq2_start =0x00000001, dq2_alu =[dq_echo()],
                  dq3_start =0x00000001, dq3_alu =[dq_echo()],
                  dq4_start =0x00000001, dq4_alu =[dq_echo()],
                  dq5_start =0x00000001, dq5_alu =[dq_echo()],
                  dq6_start =0x00000001, dq6_alu =[dq_echo()],
                  dq7_start =0x00000001, dq7_alu =[dq_echo()],
                  dq8_start =0x00000001, dq8_alu =[dq_echo()],
                  dq9_start =0x00000001, dq9_alu =[dq_echo()],
                  dq10_start=0x00000001, dq10_alu=[dq_echo()],
                  dq11_start=0x00000001, dq11_alu=[dq_echo()],
                  dq12_start=0x00000001, dq12_alu=[dq_echo()],
                  dq13_start=0x00000001, dq13_alu=[dq_echo()],
                  dq14_start=0x00000001, dq14_alu=[dq_echo()],
                  dq15_start=0x00000001, dq15_alu=[dq_echo()]
               ),
               rstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               rresp=xresp_okay
            ),
         ]),
         wait_reads()
      ]

      self.pve_vv_v2_prog = [
         #PRBS 31 with sequential addressing
         write(iters=65536, awvalid=(3,4), wvalid=(1,1), bready=(2,3), awid=[0], workers=[
            # Worker for AWID=0
            wr_worker(
               awid=0, awlen=(axibl-1), awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0,
               awaddr=addr_gen(start=0x0, lo=0x0, hi=0xFFFFFFFF, align=Awsize, alu=[incr(axibl*(self.min_awaddr_incr))]),
               wdata=data_eq_dq(
                  dq0_start =0x5A5A5A5A, dq0_alu =[dq_prbs(31)],
                  dq1_start =0xFFFFFFFF, dq1_alu =[dq_prbs(31)],
                  dq2_start =0xAAAAAAAA, dq2_alu =[dq_prbs(31)],
                  dq3_start =0x55555555, dq3_alu =[dq_prbs(31)],
                  dq4_start =0x12347890, dq4_alu =[dq_prbs(31)],
                  dq5_start =0x23458901, dq5_alu =[dq_prbs(31)],
                  dq6_start =0x89071234, dq6_alu =[dq_prbs(31)],
                  dq7_start =0xABCDE123, dq7_alu =[dq_prbs(31)],
                  dq8_start =0xFFFFAAAA, dq8_alu =[dq_prbs(31)],
                  dq9_start =0x90875431, dq9_alu =[dq_prbs(31)],
                  dq10_start=0x89898989, dq10_alu=[dq_prbs(31)],
                  dq11_start=0x90213456, dq11_alu=[dq_prbs(31)],
                  dq12_start=0x5A678901, dq12_alu=[dq_prbs(31)],
                  dq13_start=0xA5A5A5A5, dq13_alu=[dq_prbs(31)],
                  dq14_start=0xBCDEFFFF, dq14_alu=[dq_prbs(31)],
                  dq15_start=0xAAAA5555, dq15_alu=[dq_prbs(31)]
               ),
               wstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               bresp=xresp_okay
            ),
         ]),
         wait_writes(),
         read(iters=65536, arvalid=(7,8), rready=(4,5), arid=[0], workers=[
            rd_worker(
               arid=0, arlen=(axibl-1), arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0,
               araddr=addr_gen(start=0x0, lo=0x0, hi=0xFFFFFFFF, align=Arsize, alu=[incr(axibl*(self.min_araddr_incr))]),
               rdata=data_eq_dq(
                  dq0_start =0x5A5A5A5A, dq0_alu =[dq_prbs(31)],
                  dq1_start =0xFFFFFFFF, dq1_alu =[dq_prbs(31)],
                  dq2_start =0xAAAAAAAA, dq2_alu =[dq_prbs(31)],
                  dq3_start =0x55555555, dq3_alu =[dq_prbs(31)],
                  dq4_start =0x12347890, dq4_alu =[dq_prbs(31)],
                  dq5_start =0x23458901, dq5_alu =[dq_prbs(31)],
                  dq6_start =0x89071234, dq6_alu =[dq_prbs(31)],
                  dq7_start =0xABCDE123, dq7_alu =[dq_prbs(31)],
                  dq8_start =0xFFFFAAAA, dq8_alu =[dq_prbs(31)],
                  dq9_start =0x90875431, dq9_alu =[dq_prbs(31)],
                  dq10_start=0x89898989, dq10_alu=[dq_prbs(31)],
                  dq11_start=0x90213456, dq11_alu=[dq_prbs(31)],
                  dq12_start=0x5A678901, dq12_alu=[dq_prbs(31)],
                  dq13_start=0xA5A5A5A5, dq13_alu=[dq_prbs(31)],
                  dq14_start=0xBCDEFFFF, dq14_alu=[dq_prbs(31)],
                  dq15_start=0xAAAA5555, dq15_alu=[dq_prbs(31)]
               ),
               rstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               rresp=xresp_okay
            ),
         ]),
         wait_reads(),

         # DATA_EQ_ADDR
         write(iters=65536, awvalid=(1,1), wvalid=(1,1), bready=(2,3), awid=[1], workers=[
            # Worker for AWID=1
            wr_worker(
               awid=1, awlen=(axibl-1), awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0,
               awaddr=addr_gen(start=0x1000000, lo=0x1000000, hi=0xFFFFFFFF, align=Awsize, alu=[incr(axibl*(self.min_awaddr_incr))]),
               wdata=data_eq_addr(), wstrb=0xFFFFFFFF,
               bresp=xresp_okay
             )
         ]),
         wait_writes(),
         read(iters=65536, arvalid=(1,1), rready=(4,5), arid=[1], workers=[
            rd_worker(
               arid=1, arlen=(axibl-1), arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0,
               araddr=addr_gen(start=0x1000000, lo=0x1000000, hi=0xFFFFFFFF, align=Arsize, alu=[incr(axibl*(self.min_araddr_incr))]),
               rdata=data_eq_addr(), rstrb=0xFFFFFFFF, rresp=xresp_okay
            )
         ]),
         wait_reads(),

         # CLOCK Pattern
         write(iters=65536, awvalid=(3,4), wvalid=(1,1), bready=(2,3), awid=[2], workers=[
            # Worker for AWID=2
            wr_worker(
               awid=2, awlen=(axibl-1), awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0,
               awaddr=addr_gen(start=0x2000000, lo=0x2000000, hi=0xFFFFFFFF, align=Awsize, alu=[incr(axibl*(self.min_awaddr_incr))]),
               wdata=data_eq_dq(
                  dq0_start =0x55555555, dq0_alu =[dq_echo()],
                  dq1_start =0x55555555, dq1_alu =[dq_echo()],
                  dq2_start =0x55555555, dq2_alu =[dq_echo()],
                  dq3_start =0x55555555, dq3_alu =[dq_echo()],
                  dq4_start =0x55555555, dq4_alu =[dq_echo()],
                  dq5_start =0x55555555, dq5_alu =[dq_echo()],
                  dq6_start =0x55555555, dq6_alu =[dq_echo()],
                  dq7_start =0x55555555, dq7_alu =[dq_echo()],
                  dq8_start =0x55555555, dq8_alu =[dq_echo()],
                  dq9_start =0x55555555, dq9_alu =[dq_echo()],
                  dq10_start=0x55555555, dq10_alu=[dq_echo()],
                  dq11_start=0x55555555, dq11_alu=[dq_echo()],
                  dq12_start=0x55555555, dq12_alu=[dq_echo()],
                  dq13_start=0x55555555, dq13_alu=[dq_echo()],
                  dq14_start=0x55555555, dq14_alu=[dq_echo()],
                  dq15_start=0x55555555, dq15_alu=[dq_echo()]
               ),
               wstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               bresp=xresp_okay
            ),
         ]),
         wait_writes(),
         read(iters=65536, arvalid=(7,8), rready=(4,5), arid=[2], workers=[
            rd_worker(
               arid=2, arlen=(axibl-1), arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0,
               araddr=addr_gen(start=0x2000000, lo=0x2000000, hi=0xFFFFFFFF, align=Arsize, alu=[incr(axibl*(self.min_araddr_incr))]),
               rdata=data_eq_dq(
                  dq0_start =0x55555555, dq0_alu =[dq_echo()],
                  dq1_start =0x55555555, dq1_alu =[dq_echo()],
                  dq2_start =0x55555555, dq2_alu =[dq_echo()],
                  dq3_start =0x55555555, dq3_alu =[dq_echo()],
                  dq4_start =0x55555555, dq4_alu =[dq_echo()],
                  dq5_start =0x55555555, dq5_alu =[dq_echo()],
                  dq6_start =0x55555555, dq6_alu =[dq_echo()],
                  dq7_start =0x55555555, dq7_alu =[dq_echo()],
                  dq8_start =0x55555555, dq8_alu =[dq_echo()],
                  dq9_start =0x55555555, dq9_alu =[dq_echo()],
                  dq10_start=0x55555555, dq10_alu=[dq_echo()],
                  dq11_start=0x55555555, dq11_alu=[dq_echo()],
                  dq12_start=0x55555555, dq12_alu=[dq_echo()],
                  dq13_start=0x55555555, dq13_alu=[dq_echo()],
                  dq14_start=0x55555555, dq14_alu=[dq_echo()],
                  dq15_start=0x55555555, dq15_alu=[dq_echo()]
               ),
               rstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               rresp=xresp_okay
            ),
         ]),
         wait_reads(),

         # Long 1's followed by 0
         write(iters=65536, awvalid=(3,4), wvalid=(1,1), bready=(2,3), awid=[3], workers=[
            # Worker for AWID=3
            wr_worker(
               awid=3, awlen=(axibl-1), awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0,
               awaddr=addr_gen(start=0x3000000, lo=0x3000000, hi=0xFFFFFFFF, align=Awsize, alu=[incr(axibl*(self.min_awaddr_incr))]),
               wdata=data_eq_dq(
                  dq0_start =0xFEFEFEFE, dq0_alu =[dq_echo()],
                  dq1_start =0xFEFEFEFE, dq1_alu =[dq_echo()],
                  dq2_start =0xFEFEFEFE, dq2_alu =[dq_echo()],
                  dq3_start =0xFEFEFEFE, dq3_alu =[dq_echo()],
                  dq4_start =0xFEFEFEFE, dq4_alu =[dq_echo()],
                  dq5_start =0xFEFEFEFE, dq5_alu =[dq_echo()],
                  dq6_start =0xFEFEFEFE, dq6_alu =[dq_echo()],
                  dq7_start =0xFEFEFEFE, dq7_alu =[dq_echo()],
                  dq8_start =0xFEFEFEFE, dq8_alu =[dq_echo()],
                  dq9_start =0xFEFEFEFE, dq9_alu =[dq_echo()],
                  dq10_start=0xFEFEFEFE, dq10_alu=[dq_echo()],
                  dq11_start=0xFEFEFEFE, dq11_alu=[dq_echo()],
                  dq12_start=0xFEFEFEFE, dq12_alu=[dq_echo()],
                  dq13_start=0xFEFEFEFE, dq13_alu=[dq_echo()],
                  dq14_start=0xFEFEFEFE, dq14_alu=[dq_echo()],
                  dq15_start=0xFEFEFEFE, dq15_alu=[dq_echo()]
               ),
               wstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               bresp=xresp_okay
            ),
         ]),
         wait_writes(),
         read(iters=65536, arvalid=(7,8), rready=(4,5), arid=[3], workers=[
           rd_worker(
               arid=3, arlen=(axibl-1), arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0,
               araddr=addr_gen(start=0x3000000, lo=0x3000000, hi=0xFFFFFFFF, align=Arsize, alu=[incr(axibl*(self.min_araddr_incr))]),
               rdata=data_eq_dq(
                  dq0_start =0xFEFEFEFE, dq0_alu =[dq_echo()],
                  dq1_start =0xFEFEFEFE, dq1_alu =[dq_echo()],
                  dq2_start =0xFEFEFEFE, dq2_alu =[dq_echo()],
                  dq3_start =0xFEFEFEFE, dq3_alu =[dq_echo()],
                  dq4_start =0xFEFEFEFE, dq4_alu =[dq_echo()],
                  dq5_start =0xFEFEFEFE, dq5_alu =[dq_echo()],
                  dq6_start =0xFEFEFEFE, dq6_alu =[dq_echo()],
                  dq7_start =0xFEFEFEFE, dq7_alu =[dq_echo()],
                  dq8_start =0xFEFEFEFE, dq8_alu =[dq_echo()],
                  dq9_start =0xFEFEFEFE, dq9_alu =[dq_echo()],
                  dq10_start=0xFEFEFEFE, dq10_alu=[dq_echo()],
                  dq11_start=0xFEFEFEFE, dq11_alu=[dq_echo()],
                  dq12_start=0xFEFEFEFE, dq12_alu=[dq_echo()],
                  dq13_start=0xFEFEFEFE, dq13_alu=[dq_echo()],
                  dq14_start=0xFEFEFEFE, dq14_alu=[dq_echo()],
                  dq15_start=0xFEFEFEFE, dq15_alu=[dq_echo()]
               ),
               rstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               rresp=xresp_okay
            ),
         ]),
         wait_reads(),

         # Long 0's followed by 1
         write(iters=65536, awvalid=(3,4), wvalid=(1,1), bready=(2,3), awid=[4], workers=[
            # Worker for AWID=4
            wr_worker(
               awid=4, awlen=(axibl-1), awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0,
               awaddr=addr_gen(start=0x4000000, lo=0x4000000, hi=0xFFFFFFFF, align=Awsize, alu=[incr(axibl*(self.min_awaddr_incr))]),
               wdata=data_eq_dq(
                  dq0_start =0x00000001, dq0_alu =[dq_echo()],
                  dq1_start =0x00000001, dq1_alu =[dq_echo()],
                  dq2_start =0x00000001, dq2_alu =[dq_echo()],
                  dq3_start =0x00000001, dq3_alu =[dq_echo()],
                  dq4_start =0x00000001, dq4_alu =[dq_echo()],
                  dq5_start =0x00000001, dq5_alu =[dq_echo()],
                  dq6_start =0x00000001, dq6_alu =[dq_echo()],
                  dq7_start =0x00000001, dq7_alu =[dq_echo()],
                  dq8_start =0x00000001, dq8_alu =[dq_echo()],
                  dq9_start =0x00000001, dq9_alu =[dq_echo()],
                  dq10_start=0x00000001, dq10_alu=[dq_echo()],
                  dq11_start=0x00000001, dq11_alu=[dq_echo()],
                  dq12_start=0x00000001, dq12_alu=[dq_echo()],
                  dq13_start=0x00000001, dq13_alu=[dq_echo()],
                  dq14_start=0x00000001, dq14_alu=[dq_echo()],
                  dq15_start=0x00000001, dq15_alu=[dq_echo()]
               ),
               wstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               bresp=xresp_okay
            ),
         ]),
         wait_writes(),
         read(iters=65536, arvalid=(7,8), rready=(4,5), arid=[4], workers=[
           rd_worker(
               arid=4, arlen=(axibl-1), arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0,
               araddr=addr_gen(start=0x4000000, lo=0x4000000, hi=0xFFFFFFFF, align=Arsize, alu=[incr(axibl*(self.min_araddr_incr))]),
               rdata=data_eq_dq(
                  dq0_start =0x00000001, dq0_alu =[dq_echo()],
                  dq1_start =0x00000001, dq1_alu =[dq_echo()],
                  dq2_start =0x00000001, dq2_alu =[dq_echo()],
                  dq3_start =0x00000001, dq3_alu =[dq_echo()],
                  dq4_start =0x00000001, dq4_alu =[dq_echo()],
                  dq5_start =0x00000001, dq5_alu =[dq_echo()],
                  dq6_start =0x00000001, dq6_alu =[dq_echo()],
                  dq7_start =0x00000001, dq7_alu =[dq_echo()],
                  dq8_start =0x00000001, dq8_alu =[dq_echo()],
                  dq9_start =0x00000001, dq9_alu =[dq_echo()],
                  dq10_start=0x00000001, dq10_alu=[dq_echo()],
                  dq11_start=0x00000001, dq11_alu=[dq_echo()],
                  dq12_start=0x00000001, dq12_alu=[dq_echo()],
                  dq13_start=0x00000001, dq13_alu=[dq_echo()],
                  dq14_start=0x00000001, dq14_alu=[dq_echo()],
                  dq15_start=0x00000001, dq15_alu=[dq_echo()]
               ),
               rstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               rresp=xresp_okay
            ),
         ]),
         wait_reads(),

         #PRBS 31 with sequential addressing
         write(iters=65536, awvalid=(3,4), wvalid=(1,1), bready=(2,3), awid=[0], workers=[
            # Worker for AWID=0
            wr_worker(
               awid=0, awlen=(axibl-1), awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0,
               awaddr=addr_gen(start=0x6000000, lo=0x6000000, hi=0xFFFFFFFF, align=Awsize, alu=[incr(axibl*(self.min_awaddr_incr))]),
               wdata=data_eq_dq(
                  dq0_start =0x5A5A5A5A, dq0_alu =[dq_prbs(31)],
                  dq1_start =0xFFFFFFFF, dq1_alu =[dq_prbs(31)],
                  dq2_start =0xAAAAAAAA, dq2_alu =[dq_prbs(31)],
                  dq3_start =0x55555555, dq3_alu =[dq_prbs(31)],
                  dq4_start =0x12347890, dq4_alu =[dq_prbs(31)],
                  dq5_start =0x23458901, dq5_alu =[dq_prbs(31)],
                  dq6_start =0x89071234, dq6_alu =[dq_prbs(31)],
                  dq7_start =0xABCDE123, dq7_alu =[dq_prbs(31)],
                  dq8_start =0xFFFFAAAA, dq8_alu =[dq_prbs(31)],
                  dq9_start =0x90875431, dq9_alu =[dq_prbs(31)],
                  dq10_start=0x89898989, dq10_alu=[dq_prbs(31)],
                  dq11_start=0x90213456, dq11_alu=[dq_prbs(31)],
                  dq12_start=0x5A678901, dq12_alu=[dq_prbs(31)],
                  dq13_start=0xA5A5A5A5, dq13_alu=[dq_prbs(31)],
                  dq14_start=0xBCDEFFFF, dq14_alu=[dq_prbs(31)],
                  dq15_start=0xAAAA5555, dq15_alu=[dq_prbs(31)]
               ),
               wstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               bresp=xresp_okay
            ),
         ]),
         wait_writes(),
         read(iters=65536, arvalid=(7,8), rready=(4,5), arid=[0], workers=[
            rd_worker(
               arid=0, arlen=(axibl-1), arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0,
               araddr=addr_gen(start=0x6000000, lo=0x6000000, hi=0xFFFFFFFF, align=Arsize, alu=[incr(axibl*(self.min_araddr_incr))]),
               rdata=data_eq_dq(
                  dq0_start =0x5A5A5A5A, dq0_alu =[dq_prbs(31)],
                  dq1_start =0xFFFFFFFF, dq1_alu =[dq_prbs(31)],
                  dq2_start =0xAAAAAAAA, dq2_alu =[dq_prbs(31)],
                  dq3_start =0x55555555, dq3_alu =[dq_prbs(31)],
                  dq4_start =0x12347890, dq4_alu =[dq_prbs(31)],
                  dq5_start =0x23458901, dq5_alu =[dq_prbs(31)],
                  dq6_start =0x89071234, dq6_alu =[dq_prbs(31)],
                  dq7_start =0xABCDE123, dq7_alu =[dq_prbs(31)],
                  dq8_start =0xFFFFAAAA, dq8_alu =[dq_prbs(31)],
                  dq9_start =0x90875431, dq9_alu =[dq_prbs(31)],
                  dq10_start=0x89898989, dq10_alu=[dq_prbs(31)],
                  dq11_start=0x90213456, dq11_alu=[dq_prbs(31)],
                  dq12_start=0x5A678901, dq12_alu=[dq_prbs(31)],
                  dq13_start=0xA5A5A5A5, dq13_alu=[dq_prbs(31)],
                  dq14_start=0xBCDEFFFF, dq14_alu=[dq_prbs(31)],
                  dq15_start=0xAAAA5555, dq15_alu=[dq_prbs(31)]
               ),
               rstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               rresp=xresp_okay
            ),
         ]),
         wait_reads(),

         # DATA_EQ_ADDR
         write(iters=65536, awvalid=(1,1), wvalid=(1,1), bready=(2,3), awid=[1], workers=[
            # Worker for AWID=1
            wr_worker(
               awid=1, awlen=(axibl-1), awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0,
               awaddr=addr_gen(start=0x7000000, lo=0x7000000, hi=0xFFFFFFFF, align=Awsize, alu=[incr(axibl*(self.min_awaddr_incr))]),
               wdata=data_eq_addr(), wstrb=0xFFFFFFFF,
               bresp=xresp_okay
             )
         ]),
         wait_writes(),
         read(iters=65536, arvalid=(1,1), rready=(4,5), arid=[1], workers=[
            rd_worker(
               arid=1, arlen=(axibl-1), arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0,
               araddr=addr_gen(start=0x7000000, lo=0x7000000, hi=0xFFFFFFFF, align=Arsize, alu=[incr(axibl*(self.min_araddr_incr))]),
               rdata=data_eq_addr(), rstrb=0xFFFFFFFF, rresp=xresp_okay
            )
         ]),
         wait_reads(),

         # CLOCK Pattern
         write(iters=65536, awvalid=(3,4), wvalid=(1,1), bready=(2,3), awid=[2], workers=[
            # Worker for AWID=2
            wr_worker(
               awid=2, awlen=(axibl-1), awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0,
               awaddr=addr_gen(start=0x8000000, lo=0x8000000, hi=0xFFFFFFFF, align=Awsize, alu=[incr(axibl*(self.min_awaddr_incr))]),
               wdata=data_eq_dq(
                  dq0_start =0x55555555, dq0_alu =[dq_echo()],
                  dq1_start =0x55555555, dq1_alu =[dq_echo()],
                  dq2_start =0x55555555, dq2_alu =[dq_echo()],
                  dq3_start =0x55555555, dq3_alu =[dq_echo()],
                  dq4_start =0x55555555, dq4_alu =[dq_echo()],
                  dq5_start =0x55555555, dq5_alu =[dq_echo()],
                  dq6_start =0x55555555, dq6_alu =[dq_echo()],
                  dq7_start =0x55555555, dq7_alu =[dq_echo()],
                  dq8_start =0x55555555, dq8_alu =[dq_echo()],
                  dq9_start =0x55555555, dq9_alu =[dq_echo()],
                  dq10_start=0x55555555, dq10_alu=[dq_echo()],
                  dq11_start=0x55555555, dq11_alu=[dq_echo()],
                  dq12_start=0x55555555, dq12_alu=[dq_echo()],
                  dq13_start=0x55555555, dq13_alu=[dq_echo()],
                  dq14_start=0x55555555, dq14_alu=[dq_echo()],
                  dq15_start=0x55555555, dq15_alu=[dq_echo()]
               ),
               wstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               bresp=xresp_okay
            ),
         ]),
         wait_writes(),
         read(iters=65536, arvalid=(7,8), rready=(4,5), arid=[2], workers=[
            rd_worker(
               arid=2, arlen=(axibl-1), arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0,
               araddr=addr_gen(start=0x8000000, lo=0x8000000, hi=0xFFFFFFFF, align=Arsize, alu=[incr(axibl*(self.min_araddr_incr))]),
               rdata=data_eq_dq(
                  dq0_start =0x55555555, dq0_alu =[dq_echo()],
                  dq1_start =0x55555555, dq1_alu =[dq_echo()],
                  dq2_start =0x55555555, dq2_alu =[dq_echo()],
                  dq3_start =0x55555555, dq3_alu =[dq_echo()],
                  dq4_start =0x55555555, dq4_alu =[dq_echo()],
                  dq5_start =0x55555555, dq5_alu =[dq_echo()],
                  dq6_start =0x55555555, dq6_alu =[dq_echo()],
                  dq7_start =0x55555555, dq7_alu =[dq_echo()],
                  dq8_start =0x55555555, dq8_alu =[dq_echo()],
                  dq9_start =0x55555555, dq9_alu =[dq_echo()],
                  dq10_start=0x55555555, dq10_alu=[dq_echo()],
                  dq11_start=0x55555555, dq11_alu=[dq_echo()],
                  dq12_start=0x55555555, dq12_alu=[dq_echo()],
                  dq13_start=0x55555555, dq13_alu=[dq_echo()],
                  dq14_start=0x55555555, dq14_alu=[dq_echo()],
                  dq15_start=0x55555555, dq15_alu=[dq_echo()]
               ),
               rstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               rresp=xresp_okay
            ),
         ]),
         wait_reads(),

         # Long 1's followed by 0
         write(iters=65536, awvalid=(3,4), wvalid=(1,1), bready=(2,3), awid=[3], workers=[
            # Worker for AWID=3
            wr_worker(
               awid=3, awlen=(axibl-1), awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0,
               awaddr=addr_gen(start=0x9000000, lo=0x9000000, hi=0xFFFFFFFF, align=Awsize, alu=[incr(axibl*(self.min_awaddr_incr))]),
               wdata=data_eq_dq(
                  dq0_start =0xFEFEFEFE, dq0_alu =[dq_echo()],
                  dq1_start =0xFEFEFEFE, dq1_alu =[dq_echo()],
                  dq2_start =0xFEFEFEFE, dq2_alu =[dq_echo()],
                  dq3_start =0xFEFEFEFE, dq3_alu =[dq_echo()],
                  dq4_start =0xFEFEFEFE, dq4_alu =[dq_echo()],
                  dq5_start =0xFEFEFEFE, dq5_alu =[dq_echo()],
                  dq6_start =0xFEFEFEFE, dq6_alu =[dq_echo()],
                  dq7_start =0xFEFEFEFE, dq7_alu =[dq_echo()],
                  dq8_start =0xFEFEFEFE, dq8_alu =[dq_echo()],
                  dq9_start =0xFEFEFEFE, dq9_alu =[dq_echo()],
                  dq10_start=0xFEFEFEFE, dq10_alu=[dq_echo()],
                  dq11_start=0xFEFEFEFE, dq11_alu=[dq_echo()],
                  dq12_start=0xFEFEFEFE, dq12_alu=[dq_echo()],
                  dq13_start=0xFEFEFEFE, dq13_alu=[dq_echo()],
                  dq14_start=0xFEFEFEFE, dq14_alu=[dq_echo()],
                  dq15_start=0xFEFEFEFE, dq15_alu=[dq_echo()]
               ),
               wstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               bresp=xresp_okay
            ),
         ]),
         wait_writes(),
         read(iters=65536, arvalid=(7,8), rready=(4,5), arid=[3], workers=[
           rd_worker(
               arid=3, arlen=(axibl-1), arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0,
               araddr=addr_gen(start=0x9000000, lo=0x9000000, hi=0xFFFFFFFF, align=Arsize, alu=[incr(axibl*(self.min_araddr_incr))]),
               rdata=data_eq_dq(
                  dq0_start =0xFEFEFEFE, dq0_alu =[dq_echo()],
                  dq1_start =0xFEFEFEFE, dq1_alu =[dq_echo()],
                  dq2_start =0xFEFEFEFE, dq2_alu =[dq_echo()],
                  dq3_start =0xFEFEFEFE, dq3_alu =[dq_echo()],
                  dq4_start =0xFEFEFEFE, dq4_alu =[dq_echo()],
                  dq5_start =0xFEFEFEFE, dq5_alu =[dq_echo()],
                  dq6_start =0xFEFEFEFE, dq6_alu =[dq_echo()],
                  dq7_start =0xFEFEFEFE, dq7_alu =[dq_echo()],
                  dq8_start =0xFEFEFEFE, dq8_alu =[dq_echo()],
                  dq9_start =0xFEFEFEFE, dq9_alu =[dq_echo()],
                  dq10_start=0xFEFEFEFE, dq10_alu=[dq_echo()],
                  dq11_start=0xFEFEFEFE, dq11_alu=[dq_echo()],
                  dq12_start=0xFEFEFEFE, dq12_alu=[dq_echo()],
                  dq13_start=0xFEFEFEFE, dq13_alu=[dq_echo()],
                  dq14_start=0xFEFEFEFE, dq14_alu=[dq_echo()],
                  dq15_start=0xFEFEFEFE, dq15_alu=[dq_echo()]
               ),
               rstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               rresp=xresp_okay
            ),
         ]),
         wait_reads(),

         # Long 0's followed by 1
         write(iters=65536, awvalid=(3,4), wvalid=(1,1), bready=(2,3), awid=[4], workers=[
            # Worker for AWID=4
            wr_worker(
               awid=4, awlen=(axibl-1), awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0,
               awaddr=addr_gen(start=0xA000000, lo=0xA000000, hi=0xFFFFFFFF, align=Awsize, alu=[incr(axibl*(self.min_awaddr_incr))]),
               wdata=data_eq_dq(
                  dq0_start =0x00000001, dq0_alu =[dq_echo()],
                  dq1_start =0x00000001, dq1_alu =[dq_echo()],
                  dq2_start =0x00000001, dq2_alu =[dq_echo()],
                  dq3_start =0x00000001, dq3_alu =[dq_echo()],
                  dq4_start =0x00000001, dq4_alu =[dq_echo()],
                  dq5_start =0x00000001, dq5_alu =[dq_echo()],
                  dq6_start =0x00000001, dq6_alu =[dq_echo()],
                  dq7_start =0x00000001, dq7_alu =[dq_echo()],
                  dq8_start =0x00000001, dq8_alu =[dq_echo()],
                  dq9_start =0x00000001, dq9_alu =[dq_echo()],
                  dq10_start=0x00000001, dq10_alu=[dq_echo()],
                  dq11_start=0x00000001, dq11_alu=[dq_echo()],
                  dq12_start=0x00000001, dq12_alu=[dq_echo()],
                  dq13_start=0x00000001, dq13_alu=[dq_echo()],
                  dq14_start=0x00000001, dq14_alu=[dq_echo()],
                  dq15_start=0x00000001, dq15_alu=[dq_echo()]
               ),
               wstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               bresp=xresp_okay
            ),
         ]),
         wait_writes(),
         read(iters=65536, arvalid=(7,8), rready=(4,5), arid=[4], workers=[
            rd_worker(
               arid=4, arlen=(axibl-1), arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0,
               araddr=addr_gen(start=0xA000000, lo=0xA000000, hi=0xFFFFFFFF, align=Arsize, alu=[incr(axibl*(self.min_araddr_incr))]),
               rdata=data_eq_dq(
                  dq0_start =0x00000001, dq0_alu =[dq_echo()],
                  dq1_start =0x00000001, dq1_alu =[dq_echo()],
                  dq2_start =0x00000001, dq2_alu =[dq_echo()],
                  dq3_start =0x00000001, dq3_alu =[dq_echo()],
                  dq4_start =0x00000001, dq4_alu =[dq_echo()],
                  dq5_start =0x00000001, dq5_alu =[dq_echo()],
                  dq6_start =0x00000001, dq6_alu =[dq_echo()],
                  dq7_start =0x00000001, dq7_alu =[dq_echo()],
                  dq8_start =0x00000001, dq8_alu =[dq_echo()],
                  dq9_start =0x00000001, dq9_alu =[dq_echo()],
                  dq10_start=0x00000001, dq10_alu=[dq_echo()],
                  dq11_start=0x00000001, dq11_alu=[dq_echo()],
                  dq12_start=0x00000001, dq12_alu=[dq_echo()],
                  dq13_start=0x00000001, dq13_alu=[dq_echo()],
                  dq14_start=0x00000001, dq14_alu=[dq_echo()],
                  dq15_start=0x00000001, dq15_alu=[dq_echo()]
               ),
               rstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               rresp=xresp_okay
            ),
         ]),
         wait_reads(),

         #PRBS 31 with random addressing
         write(iters=65536, awvalid=(3,4), wvalid=(1,1), bready=(2,3), awid=[0], workers=[
            # Worker for AWID=0
            wr_worker(
               awid=0, awlen=(axibl-1), awsize=self.max_awsize, awburst=axburst_incr, awlock=0, awcache=0, awprot=0, awqos=0, awregion=0, awuser=0,
               awaddr=addr_gen(start=0x0000000, lo=0x0000000, hi=0xFFFFFFFF, align=Awaddr_align, alu=[rand(32)]),
               wdata=data_eq_dq(
                  dq0_start =0x5A5A5A5A, dq0_alu =[dq_prbs(31)],
                  dq1_start =0xFFFFFFFF, dq1_alu =[dq_prbs(31)],
                  dq2_start =0xAAAAAAAA, dq2_alu =[dq_prbs(31)],
                  dq3_start =0x55555555, dq3_alu =[dq_prbs(31)],
                  dq4_start =0x12347890, dq4_alu =[dq_prbs(31)],
                  dq5_start =0x23458901, dq5_alu =[dq_prbs(31)],
                  dq6_start =0x89071234, dq6_alu =[dq_prbs(31)],
                  dq7_start =0xABCDE123, dq7_alu =[dq_prbs(31)],
                  dq8_start =0xFFFFAAAA, dq8_alu =[dq_prbs(31)],
                  dq9_start =0x90875431, dq9_alu =[dq_prbs(31)],
                  dq10_start=0x89898989, dq10_alu=[dq_prbs(31)],
                  dq11_start=0x90213456, dq11_alu=[dq_prbs(31)],
                  dq12_start=0x5A678901, dq12_alu=[dq_prbs(31)],
                  dq13_start=0xA5A5A5A5, dq13_alu=[dq_prbs(31)],
                  dq14_start=0xBCDEFFFF, dq14_alu=[dq_prbs(31)],
                  dq15_start=0xAAAA5555, dq15_alu=[dq_prbs(31)]
               ),
               wstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               bresp=xresp_okay
            ),
         ]),
         wait_writes(),
         read(iters=65536, arvalid=(7,8), rready=(4,5), arid=[0], workers=[
            rd_worker(
               arid=0, arlen=(axibl-1), arsize=self.max_arsize, arburst=axburst_incr, arlock=0, arcache=0, arprot=0, arqos=0, arregion=0, aruser=0,
               araddr=addr_gen(start=0x0000000, lo=0x0000000, hi=0xFFFFFFFF, align=Araddr_align, alu=[rand(32)]),
               rdata=data_eq_dq(
                  dq0_start =0x5A5A5A5A, dq0_alu =[dq_prbs(31)],
                  dq1_start =0xFFFFFFFF, dq1_alu =[dq_prbs(31)],
                  dq2_start =0xAAAAAAAA, dq2_alu =[dq_prbs(31)],
                  dq3_start =0x55555555, dq3_alu =[dq_prbs(31)],
                  dq4_start =0x12347890, dq4_alu =[dq_prbs(31)],
                  dq5_start =0x23458901, dq5_alu =[dq_prbs(31)],
                  dq6_start =0x89071234, dq6_alu =[dq_prbs(31)],
                  dq7_start =0xABCDE123, dq7_alu =[dq_prbs(31)],
                  dq8_start =0xFFFFAAAA, dq8_alu =[dq_prbs(31)],
                  dq9_start =0x90875431, dq9_alu =[dq_prbs(31)],
                  dq10_start=0x89898989, dq10_alu=[dq_prbs(31)],
                  dq11_start=0x90213456, dq11_alu=[dq_prbs(31)],
                  dq12_start=0x5A678901, dq12_alu=[dq_prbs(31)],
                  dq13_start=0xA5A5A5A5, dq13_alu=[dq_prbs(31)],
                  dq14_start=0xBCDEFFFF, dq14_alu=[dq_prbs(31)],
                  dq15_start=0xAAAA5555, dq15_alu=[dq_prbs(31)]
               ),
               rstrb=strb_eq_dm(
                  dm0_start =0xFFFFFFFF, dm0_alu =[dm_echo()],
                  dm1_start =0xFFFFFFFF, dm1_alu =[dm_echo()],
                  dm2_start =0xFFFFFFFF, dm2_alu =[dm_echo()],
                  dm3_start =0xFFFFFFFF, dm3_alu =[dm_echo()],
                  dm4_start =0xFFFFFFFF, dm4_alu =[dm_echo()],
                  dm5_start =0xFFFFFFFF, dm5_alu =[dm_echo()],
                  dm6_start =0xFFFFFFFF, dm6_alu =[dm_echo()],
                  dm7_start =0xFFFFFFFF, dm7_alu =[dm_echo()]
               ),
               rresp=xresp_okay
            ),
         ]),
         wait_reads()
     ]

      ###########################################################
      # HBM2E kernels
      ###########################################################

      # TEST HBM IP
      #HBM IP: define i_BL according to pBL8 channel enabled or not, enabled -- i_BL= 2(64B), disabled --i_BL= 1(32B) ,
      hbm_ip_config = None
      for ip_name in config['system']['ips']:
          if config['system']['ips'][ip_name]['type'] == "hbm_fp":
              hbm_ip_config = config['system']['ips'][ip_name]

      ADDR_INDEX1 = 0x00000000
      ADDR_INDEX2 = 0x3fffffff
      max_size      = min(self.max_awsize, self.max_awsize)

      if hbm_ip_config is None:
         i_BL   = 1  if (max_size >= 6) else 2
         NUM_Tra = "Sequential_Medium"
      else:
         hbm_params = hbm_ip_config['parameters']
         HBM_CH_ID  = 0
         enabled_ch = 0
         for i in range(8):
            if hbm_params[f'CTRL_CH{i}_EN']['value']=="true":
               if enabled_ch == driver_index:
                  is_clone  = hbm_params[f'CTRL_CH{i}_CLONE_OF_ID']['value']
                  HBM_CH_ID = i if is_clone=="None" else int(is_clone)
                  break
               enabled_ch += 1

         fabric_NoC_enabled = hbm_params[f'EX_DESIGN_FABRIC_NOC_MODE']['value'] != "FABRIC_NOC_MODE_NONE"
         is_pBL8 = hbm_params[f'CTRL_CH{HBM_CH_ID}_PSEUDO_BL8_EN']['value'].lower() == "true"
         i_BL   = 2 if ((is_pBL8 or fabric_NoC_enabled) and not (max_size >= 6)) else 1
         NUM_Tra = hbm_params[f'EX_DESIGN_DEFAULT_TRAFFIC_PATTERN']['value']

         if hbm_params['SYS_INFO_DEVICE']['value'][3] == 'G' or hbm_params['SYS_INFO_DEVICE']['value'][3] == 'E':
            ADDR_INDEX2 = 0x1fffffff

      min_addr_incr = max(self.min_awaddr_incr, self.min_araddr_incr)


      min_addr_incr1 = min_addr_incr * i_BL

      # define test duration based on EX_DESIGN_DEFAULT_TRAFFIC_PATTERN parameter
      if re.match("Sequential_Short", NUM_Tra):
          ITERS = 16 
      elif re.match("Sequential_Long", NUM_Tra):
          ITERS = 3200 
      else :
          ITERS = 1600 

      ALIGN_SIZE = max_size + 1
      start_addr_0_0 = ADDR_INDEX1 + ((ADDR_INDEX2-ADDR_INDEX1+1)//4) * 0
      start_addr_0_1 = ADDR_INDEX1 + ((ADDR_INDEX2-ADDR_INDEX1+1)//4) * 1 - min_addr_incr1 * ITERS
      start_addr_1_0 = ADDR_INDEX1 + ((ADDR_INDEX2-ADDR_INDEX1+1)//4) * 1
      start_addr_1_1 = ADDR_INDEX1 + ((ADDR_INDEX2-ADDR_INDEX1+1)//4) * 2 - min_addr_incr1 * ITERS
      start_addr_2_0 = ADDR_INDEX1 + ((ADDR_INDEX2-ADDR_INDEX1+1)//4) * 2
      start_addr_2_1 = ADDR_INDEX1 + ((ADDR_INDEX2-ADDR_INDEX1+1)//4) * 3 - min_addr_incr1 * ITERS
      start_addr_3_0 = ADDR_INDEX1 + ((ADDR_INDEX2-ADDR_INDEX1+1)//4) * 3
      start_addr_3_1 = ADDR_INDEX1 + ((ADDR_INDEX2-ADDR_INDEX1+1)//4) * 4 - min_addr_incr1 * ITERS

      # TEST HBM TRAFFIC PATTERN
      # HBM traffic pattern test for defined space
      self.hbm_simulation1 = [
     	   *self.gencmd('wr', iters=ITERS, bl=i_BL, addr_range=(ADDR_INDEX1, ADDR_INDEX2), data="prbs7"),
     	   wait_writes(),
     	   sleep(256),
     	   *self.gencmd('rd', iters=ITERS, bl=i_BL, addr_range=(ADDR_INDEX1, ADDR_INDEX2), data="prbs7"),
     	   wait_reads(),
      ]


      # Order for block burst lengths i_BL*: 1,2,4,8,16,32,64,1
      # In case of greater than 512 data width default to BL1 in all cases
      hbm_bl = [1, 2, 4, 8, 16, 32, 64, 1] if (max_size < 7) else [1, 1, 1, 1, 1, 1, 1, 1]

      # HBM traffic pattern test for 1GB memeory space
      self.hbm_simulation = [
         *self.gencmd('wr', iters=max(1, math.floor(ITERS/hbm_bl[0])), bl=i_BL*hbm_bl[0], addr_range=(start_addr_0_0, ADDR_INDEX2), data="prbs7"),
         wait_writes(),
         sleep(256),

         parallel([
            *self.gencmd('rd', iters=max(1, math.floor(ITERS/hbm_bl[0])), bl=i_BL*hbm_bl[0], addr_range=(start_addr_0_0, ADDR_INDEX2), data="prbs7"),
            *self.gencmd('wr', iters=max(1, math.floor(ITERS/hbm_bl[1])), bl=i_BL*hbm_bl[1], addr_range=(start_addr_0_1, ADDR_INDEX2), data="prbs7"),
         ]),
         wait_reads(),
         wait_writes(),
         sleep(256),

         parallel([
            *self.gencmd('rd', iters=max(1, math.floor(ITERS/hbm_bl[1])), bl=i_BL*hbm_bl[1], addr_range=(start_addr_0_1, ADDR_INDEX2), data="prbs7"),
            *self.gencmd('wr', iters=max(1, math.floor(ITERS/hbm_bl[2])), bl=i_BL*hbm_bl[2], addr_range=(start_addr_1_0, ADDR_INDEX2), data="prbs7"),
         ]),
         wait_reads(),
         wait_writes(),
         sleep(256),

         parallel([
            *self.gencmd('rd', iters=max(1, math.floor(ITERS/hbm_bl[2])), bl=i_BL*hbm_bl[2], addr_range=(start_addr_1_0, ADDR_INDEX2), data="prbs7"),
            *self.gencmd('wr', iters=max(1, math.floor(ITERS/hbm_bl[3])), bl=i_BL*hbm_bl[3], addr_range=(start_addr_1_1, ADDR_INDEX2), data="prbs7"),
         ]),
         wait_reads(),
         wait_writes(),
         sleep(256),

         parallel([
            *self.gencmd('rd', iters=max(1, math.floor(ITERS/hbm_bl[3])), bl=i_BL*hbm_bl[3], addr_range=(start_addr_1_1, ADDR_INDEX2), data="prbs7"),
            *self.gencmd('wr', iters=max(1, math.floor(ITERS/hbm_bl[4])), bl=i_BL*hbm_bl[4], addr_range=(start_addr_2_0, ADDR_INDEX2), data="prbs7"),
         ]),
         wait_reads(),
         wait_writes(),
         sleep(256),

         parallel([
            *self.gencmd('rd', iters=max(1, math.floor(ITERS/hbm_bl[4])), bl=i_BL*hbm_bl[4], addr_range=(start_addr_2_0, ADDR_INDEX2), data="prbs7"),
            *self.gencmd('wr', iters=max(1, math.floor(ITERS/hbm_bl[5])), bl=i_BL*hbm_bl[5], addr_range=(start_addr_2_1, ADDR_INDEX2), data="prbs7"),
         ]),
         wait_reads(),
         wait_writes(),
         sleep(256),

         parallel([
            *self.gencmd('rd', iters=max(1, math.floor(ITERS/hbm_bl[5])), bl=i_BL*hbm_bl[5], addr_range=(start_addr_2_1, ADDR_INDEX2), data="prbs7"),
            *self.gencmd('wr', iters=max(1, math.floor(ITERS/hbm_bl[6])), bl=i_BL*hbm_bl[6], addr_range=(start_addr_3_0, ADDR_INDEX2), data="prbs7"),
         ]),
         wait_reads(),
         wait_writes(),
         sleep(256),

         parallel([
            *self.gencmd('rd', iters=max(1, math.floor(ITERS/hbm_bl[6])), bl=i_BL*hbm_bl[6], addr_range=(start_addr_3_0, ADDR_INDEX2), data="prbs7"),
            *self.gencmd('wr', iters=max(1, math.floor(ITERS/hbm_bl[7])), bl=i_BL*hbm_bl[7], addr_range=(start_addr_3_1, ADDR_INDEX2), data="prbs7"),
         ]),
         wait_reads(),
         wait_writes(),
         sleep(256),

         *self.gencmd('rd', iters=max(1, math.floor(ITERS/hbm_bl[7])), bl=i_BL*hbm_bl[7], addr_range=(start_addr_3_1, ADDR_INDEX2), data="prbs7"),
         wait_reads(),
      ]

      ###########################################################
      # Memory Subsystem kernels
      ###########################################################

      # MemSS out-of-the box traffic pattern
      self.memss_default = [
         # Single write/read
         *self.gencmd("wr"),
         wait_writes(),
         *self.gencmd("rd"),
         wait_reads(),

         sleep(100),

         # Bulk write/read
         *self.gencmd("wr", iters=256),
         wait_writes(),
         *self.gencmd("rd", iters=256),
         wait_reads(),

         sleep(100),

         # Bulk write/read with random access
         *self.gencmd("wr", iters=256, addr="rand"),
         wait_writes(),
         *self.gencmd("rd", iters=256, addr="rand"),
         wait_reads(),

         sleep(100),

         # Bulk write/read with random access and burst
         *self.gencmd("wr", iters=256, addr="rand", bl=2, data="prbs7"),
         wait_writes(),
         *self.gencmd("rd", iters=256, addr="rand", bl=2, data="prbs7"),
         wait_reads()


      ]


   def gencmd(
      self,
      cmd:        str,
      iters:      int = 1,
      addr:       Union[int, List[int], str, List[str]] = "seq",
      addr_range: Union[None, Tuple[int, int]] = None,
      data:       Union[None, int, List[int], str, List[str]] = "addr",
      strb:       Union[int, List[int], str, List[str]] = -1,   # -1 translates to all 1s in binary
      bl:         Union[int, List[int]] = 1,
      size:       Union[float, List[float]] = 1.0,
      burst:      Union[int, List[int]] = axburst_incr,
      lock:       Union[int, List[int]] = 0,
      cache:      Union[int, List[int]] = 0,
      prot:       Union[int, List[int]] = 0,
      qos:        Union[int, List[int]] = 0,
      region:     Union[int, List[int]] = 0,
      auser:      Union[int, List[int]] = 0,
      resp:       Union[int, List[int]] = xresp_okay,
      id:         Union[int, List[int]] = 0,
      addr_bw:    Union[float, Tuple[int, int]] = 1.0,
      data_bw:    Union[float, Tuple[int, int]] = 1.0,
      resp_bw:    Union[float, Tuple[int, int]] = 1.0,
      resume:     Union[bool, List[bool]] = False,
      inspect:    bool = True):
      '''
      Helper function to construct a write or read instruction for the Memory AXI Driver.
      By default, the write/read instruction generates a single AXI transaction that accesses sequential addresses with data matching the address.
      This default behavior can be customized by overriding the appropriate input arguments to this function.

         Parameters:
            cmd:        Supported values are:
                           "wr":          Generate AXI write transactions
                           "rd":          Generate AXI read transactions
            iters:      Number of AXI write/read transactions to generate.

            addr:       Integer  -> AxADDR value
                        String   -> Preset for AxADDR pattern. Supported values are:
                           "seq[_strideX][_byteen]":  Sequential access. Optional modifiers:
                                                      _strideX -> X must be an integer, and will be multiplied to address increment. Ex: _stride2 multiplies increment by 2.
                                                      _byteen -> Repeats every address twice. Can be used for byte-enable testing with bl=1.
                           "rand[_byteen]":           Random access. Optional modifiers:
                                                      _byteen -> Repeats every address twice. Can be used for byte-enable testing with bl=1.
            addr_range: Address range formatted as a tuple (lo_addr, hi_addr), where lo_addr <= AxADDR <= hi_addr. Ex: (0x0000, 0x0FFF).
                        By default the address range matches the full address space derived based on address port width.
                        addr_range is only applicable if addr preset is used.

            data:       None     -> Skip error checking for RDATA (only applicable for 'read' command)
                        Integer  -> WDATA value for 'write' transaction, expected RDATA value for 'read' transaction
                        String   -> Preset for xDATA pattern. Supported values are:
                           "addr":        xDATA matches the address of the corresponding command
                           "id":          xDATA matches the AxID of the corresponding command
                           "clock":       Every DQ pin toggles between 1 and 0 on every clock cycle
                           "pulse0":      Every DQ pin outputs a long train of 1s with a single intermittent 0
                           "pulse1":      Every DQ pin outputs a long train of 0s with a single intermittent 1
                           "walking0":    Each DQ pin within a byte lane takes turns outputting a single 0. Useful for finding crosstalk issues and normally used for simultaneous switching noise (SSN) testing. The data bit (DQ pin) that has a different value from the rest will be more likely to be affected by the rest of the data bits, especially the adjacent bits.
                           "walking1":    Each DQ pin within a byte lane takes turns outputting a single 1. Useful for finding crosstalk issues and normally used for simultaneous switching noise (SSN) testing. The data bit (DQ pin) that has a different value from the rest will be more likely to be affected by the rest of the data bits, especially the adjacent bits.
                           "prbs7":       Every DQ pin outputs a PRBS7 pattern initialized with a unique seed.
                           "prbs15":      Every DQ pin outputs a PRBS15 pattern initialized with a unique seed.
                           "prbs23":      Every DQ pin outputs a PRBS23 pattern initialized with a unique seed.
                           "prbs31":      Every DQ pin outputs a PRBS31 pattern initialized with a unique seed.

            strb:       Integer  -> WSRTB value for 'write' transaction, bytemask for error checking RDATA for 'read' transaction
                        String   -> Preset for xSTRB pattern. Supported values are:
                           "prbs7":       Every DM pin outputs a PRBS7 pattern initialized with a unique seed.
                           "prbs15":      Every DM pin outputs a PRBS15 pattern initialized with a unique seed.
                           "prbs23":      Every DM pin outputs a PRBS23 pattern initialized with a unique seed.
                           "prbs31":      Every DM pin outputs a PRBS31 pattern initialized with a unique seed.
                           "byteen_test": Continually generate a random xSTRB followed by its inverse. Can be used for byte-enable testing with bl=1.

            bl:         Burst length value. AxLEN is derived based on this (AxLEN = bl - 1)
            size:       Fraction of data width to transfer. Ex: 1, 0.5, 0.25, 0.125, ...
                        AxSIZE is derived based on this.
            burst:      AxBURST value
            lock:       AxLOCK value
            cache:      AxCACHE value
            prot:       AxPROT value
            qos:        AxQOS value
            region:     AxREGION value
            auser:      AxUSER value
            resp:       None     -> Skip error checking for xRESP
                        Integer  -> Expected xRESP value. If expected value is an error response RDATA checking will be disabled.

            id:         Integer  -> AxID value
                        List     -> Number of unique AxIDs in this list determines number of workers. And the list itself represents the AxID sequence.
                                    NOTE: Each AxID is represented by a unique "worker", which can generate a distinct traffic pattern for that specific AxID.
                                    The address range is evenly split among all workers.
                                    Ex: addr_range 0x00000000-0x3fffffff split among 4 workers is 0x00000000-0x0fffffff, 0x10000000-0x1fffffff, 0x20000000-0x2fffffff, 0x30000000-0x3fffffff
                                    You can customize the traffic stream for each worker by providing a list for supported args.
                                    Ex: bl=4 will use burst length 4 for all workers, bl=[2, 4] will use burst lengths 2 and 4 for the first and second workers respectively.
                                    The following args can be customized per worker by providing a list:
                                    addr, data, strb, bl, size, burst, lock, cache, prot, qos, region, auser, resp, resume

            addr_bw:    AW/AR channel bandwidth specified using duty-cycle.
                        Float    -> Duty-cycle specified as percentage. Ex: 0.5 for 50% duty-cycle
                        Tuple    -> Duty-cycle specified as (pulse width, period). Ex: (1, 2) or (4, 8)
            data_bw:    W channel bandwidth specified using duty-cycle.
                        Float    -> Duty-cycle specified as percentage. Ex: 0.5 for 50% duty-cycle
                        Tuple    -> Duty-cycle specified as (pulse width, period). Ex: (1, 2) or (4, 8)
            resp_bw:    B/R channel bandwidth specified using duty-cycle.
                        Float    -> Duty-cycle specified as percentage. Ex: 0.5 for 50% duty-cycle
                        Tuple    -> Duty-cycle specified as (pulse width, period). Ex: (1, 2) or (4, 8)

            resume:     Continue address/data/strb pattern preset from previous command. If resume=True, addr, data, and strb values must match previous command.

            inspect:    Scan the traffic pattern for potential issues and throw errors. Set inspect=False to print warnings instead.

         Returns:
            List of write or read instruction(s) for the Memory AXI Driver's kernel.
            NOTE: you can use Python's splat operator (*) to unpack the list for insertion into a larger list. Ex: my_kernel = [*gencmd(...), *gencmd(...), ...]
      '''

      # Helper function to throw errors/warnings
      def print_message(is_error, message):
         if is_error:
            raise Exception(message)
         else:
            print(f"Warning: {message}")

      if cmd not in ['wr', 'rd']:
         raise ValueError(f'Unrecognized cmd "{cmd}"')

      max_addr      = self.max_awaddr if cmd == 'wr' else self.max_araddr
      min_addr_incr = self.min_awaddr_incr if cmd == 'wr' else self.min_araddr_incr
      max_size      = self.max_awsize if cmd == 'wr' else self.max_arsize

      # Allowed address range for this read/write command
      min_addr = 0
      if addr_range is not None:
         min_addr = addr_range[0]
         max_addr = addr_range[1]

      # Get list of unique AxIDs
      if not isinstance(id, list):
         id = [id]
      ids = sorted(list(set(id)))

      # If only a single value is provided for the below fields, replicate that for all workers
      if not isinstance(addr,    list): addr    = [addr]    * len(ids)
      if not isinstance(data,    list): data    = [data]    * len(ids)
      if not isinstance(strb,    list): strb    = [strb]    * len(ids)
      if not isinstance(bl,      list): bl      = [bl]      * len(ids)
      if not isinstance(size,    list): size    = [size]    * len(ids)
      if not isinstance(burst,   list): burst   = [burst]   * len(ids)
      if not isinstance(lock,    list): lock    = [lock]    * len(ids)
      if not isinstance(cache,   list): cache   = [cache]   * len(ids)
      if not isinstance(prot,    list): prot    = [prot]    * len(ids)
      if not isinstance(qos,     list): qos     = [qos]     * len(ids)
      if not isinstance(region,  list): region  = [region]  * len(ids)
      if not isinstance(auser,   list): auser   = [auser]   * len(ids)
      if not isinstance(resp,    list): resp    = [resp]    * len(ids)
      if not isinstance(resume,  list): resume  = [resume]  * len(ids)

      # The below loop creates the traffic pattern for each AxID by configuring the appropriate worker
      workers = []
      for worker_num, (id_i, addr_i, data_i, strb_i, bl_i, size_i, burst_i, lock_i, cache_i, prot_i, qos_i, region_i, auser_i, resp_i, resume_i) in \
         enumerate(zip(ids,  addr,   data,   strb,   bl,   size,   burst,   lock,   cache,   prot,   qos,   region,   auser,   resp,   resume)):
         # Some AXI port values
         axid     = id_i
         axlen    = bl_i - 1
         axsize   = max_size + self.clog2(size_i)   # = clog2(data_width * size / 8) = clog2(data_width / 8) + clog2(size) = max_size + clog2(size)
         axburst  = burst_i
         axlock   = lock_i
         axcache  = cache_i
         axprot   = prot_i
         axqos    = qos_i
         axregion = region_i
         axuser   = auser_i
         xresp    = resp_i

         # AXI4 spec restricts the allowed burst lengths for wrapping bursts
         axburst_wrap_bl_list = [2, 4, 8, 16]
         if axburst == axburst_wrap and bl_i not in axburst_wrap_bl_list:
            print_message(inspect, f"AXI4 spec violation! Burst length {bl_i} is not permitted for wrapping bursts. Permitted burst lengths are {axburst_wrap_bl_list}.")

         # AXI4 spec restricts that bursts shouldn't cross a 4KB boundary
         if axburst != axburst_fixed and bl_i * (2 ** axsize) > 4 * 1024:
            print_message(inspect, f"AXI4 spec violation! Burst length {bl_i} and AxSIZE={axsize} crosses 4KB boundary. Try lowering burst length or AxSIZE.")

         # Byte-enable presets don't work with burst access
         if isinstance(addr_i, str) and "byteen" in addr_i and bl_i > 1:
            print_message(inspect, f"AxADDR pattern preset '{addr_i}' is intended for byte-enable testing and only works with burst length 1.")
         if isinstance(strb_i, str) and "byteen" in strb_i and bl_i > 1:
            print_message(inspect, f"xSTRB pattern preset '{strb_i}' is intended for byte-enable testing and only works with burst length 1.")

         # Force disable RDATA error checking if expected RRESP is an error response
         if xresp in [xresp_slverr, xresp_decerr]:
            strb = 0

         # Number of byte-address locations being accessed by a single AXI command
         if axburst == axburst_fixed:
            num_byte_addrs = (2 ** axsize)
         else:
            num_byte_addrs = (2 ** axsize) * bl_i

         # Address range for this worker:
         # Evenly split address space among workers
         # Ex: addr range 0x00000000-0x3fffffff split among 4 workers is 0x00000000-0x0fffffff, 0x10000000-0x1fffffff, 0x20000000-0x2fffffff, 0x30000000-0x3fffffff
         lo_addr = min_addr + worker_num     * ((max_addr+1 - min_addr) // len(ids))
         hi_addr = min_addr + (worker_num+1) * ((max_addr+1 - min_addr) // len(ids)) - 1

         # Align address bounds to transfer size. This avoids problems such as crossing 4KB boundaries and bursts exceeding the address range.
         aligned_lo_addr = self.roundup(lo_addr, num_byte_addrs)
         aligned_hi_addr = self.rounddown(hi_addr+1, num_byte_addrs) - 1
         if (lo_addr != aligned_lo_addr) or (hi_addr != aligned_hi_addr):
            print_message(False, f"Address Range has been adjusted for Worker ID={hex(id_i)} for transfer size alignment. Before:({hex(lo_addr)},{hex(hi_addr)}), After:({hex(aligned_lo_addr)},{hex(aligned_hi_addr)})")
         if aligned_lo_addr <= aligned_hi_addr:
            lo_addr = aligned_lo_addr
            hi_addr = aligned_hi_addr
         else:
            print_message(inspect, f"Unable to align address bounds to burst size. Try expanding the address range.")

         # Start from the lower address bound.
         # We'll align the start address later after we parse the address preset.
         start_addr = lo_addr

         # Calculate Worker Iteration count, used for Address Retreading Warning Check
         worker_iters = id.count(id_i) * (iters // len(id)) + id[ : iters % len(id)].count(id_i)

         # Address pattern
         axaddr = None
         # Accessing a single address
         if isinstance(addr_i, int):
            axaddr = addr_i
         # Address pattern preset
         elif isinstance(addr_i, str):
            addr_alu = None
            # Sequential access
            # Ex: seq, seq_stride16_byteen
            m = re.match(r"^seq(_stride\d+)?(_byteen)?$", addr_i)
            if m:
               # Parse the args
               stride, byteen, = m.groups()
               stride = 1 if stride is None else int( stride[len("_stride"):] )  # stride = "_stride16"  -->  stride = 16
               byteen = False if byteen is None else True
               # Increment by number of byte-address locations accessed in a single read/write command
               addr_alu   = [incr(num_byte_addrs * stride)]
               # Align address to data word size.
               #
               # align = log2(burst_size) = AxSIZE
               addr_align = axsize
               # For byte-enable test mode, repeat every address twice.
               # Do this by injecting echo() in between all the ALU ops. Ex: addr_alu=[opA, opB, opC] --> addr_alu=[echo, opA, echo, opB, echo, opC]
               if byteen:
                  for i in range(0, len(addr_alu)*2, 2):
                     addr_alu.insert(i, echo())
               # Address Retreading Check
               # Determine real address range based on Address ALU Arg Width
               addr_range_i = hi_addr+1 - lo_addr
               max_offset = ( (1 << self.addr_alu_arg_width) - 1 ) << addr_align
               # If the max offset supported by the Address ALU is smaller than the address range, reduce range to max offset
               if addr_range_i > max_offset:
                  addr_range_i = max_offset
               # Calculate the data range: Real Address range + bytes covered by the highest address in range
               data_range = addr_range_i + num_byte_addrs - 1
               # Calculate number of unique addresses within the data range.
               num_unique_addr = data_range // (num_byte_addrs * stride)
               # Byte-enable test repeats every address twice, hence double the number of safe iterations
               worker_iters_max_no_retread = 2 * num_unique_addr if byteen else num_unique_addr
               if worker_iters > worker_iters_max_no_retread:
                  # Calculate the safe cmd iters to avoid retreading for this worker ID
                  # cmd_iters_max_no_retread = Command Iteration of the (n+1)th occurrence of the Worker ID minus 1, where n=worker_iters_max_no_retread
                  #                          = Number of full cycles through ID list                        + Index of the (n+1)th occurrence (this includes the minus 1)
                  cmd_iters_max_no_retread = ( (worker_iters_max_no_retread) // id.count(id_i)) * len(id) + [i for i, n in enumerate(id) if n == id_i][(worker_iters_max_no_retread % id.count(id_i))]
                  print_message(inspect, f"Address Retreading detected for Worker ID {hex(id_i)}! To avoid address repetition please use a safe value for iters: current={iters} safe={cmd_iters_max_no_retread}")
            # Random access
            # Ex: rand, rand_byteen
            m = re.match(r"^rand(_byteen)?$", addr_i)
            if m:
               # Parse the args
               byteen, = m.groups()
               byteen = False if byteen is None else True
               # LFSR width is derived based on address range (so the LFSR doesn't output numbers that
               # are larger than the address span)
               #
               # Memory AXI Driver's address bounds check for:  lo <= addr <= hi
               # n-bit LFSR generates numbers from: 0 <= lfsr output <= (2^n - 1)
               # So the output address range is:    lo <= addr <= lo + (2^n - 1)
               # Hence to avoid exceeding the address bound: lo + (2^n - 1) <= hi  -->  n = floor(log2(hi - lo + 1))
               #
               # Consequently this implies that using non-power-of-2 address span with rand pattern
               # will not cover all addresses
               #
               # To account for the aligned address bounds, add the transfer size to the calculation of the LFSR Width.
               # This accounts for the data range covered by the random address generator
               lfsr_width = math.floor(math.log(hi_addr - lo_addr + num_byte_addrs) / math.log(2))
               # Align address to data word size and burst length. We align address to burst length
               # to ensure the addr bits corresponding to burst index are not randomized by LFSR,
               # since that could lead to stomping on previously accessed addresses.
               # Consequently this implies that using non-power-of-2 burst length with rand pattern
               # will not cover all addresses.
               #
               # align = log2(burst_size * burst_length)
               addr_align = self.clog2(num_byte_addrs)
               # The effective LSFR width (LFSR width - align) cannot excceed the Address ALU Arg Width
               if lfsr_width - addr_align > self.addr_alu_arg_width:
                  lfsr_width = self.addr_alu_arg_width + addr_align
               addr_alu   = [rand(lfsr_width)]
               # For byte-enable test mode, repeat every address twice.
               # Do this by injecting echo() in between all the ALU ops. Ex: addr_alu=[opA, opB, opC] --> addr_alu=[echo, opA, echo, opB, echo, opC]
               if byteen:
                  for i in range(0, len(addr_alu)*2, 2):
                     addr_alu.insert(i, echo())
               # Address Retreading Check
               # Calculate number of unique addresses within the address bounds
               num_unique_addr = 2 ** (lfsr_width - addr_align)
               # Byte-enable test repeats every address twice, hence double the number of safe iterations
               worker_iters_max_no_retread = 2 * num_unique_addr if byteen else num_unique_addr
               if worker_iters > worker_iters_max_no_retread:
                  # Calculate the safe cmd iters to avoid retreading for this worker ID
                  # cmd_iters_max_no_retread = Command Iteration of the (n+1)th occurrence of the Worker ID minus 1, where n=worker_iters_max_no_retread
                  #                          = Number of full cycles through ID list                        + Index of the (n+1)th occurrence (this includes the minus 1)
                  cmd_iters_max_no_retread = ( (worker_iters_max_no_retread) // id.count(id_i)) * len(id) + [i for i, n in enumerate(id) if n == id_i][(worker_iters_max_no_retread % id.count(id_i))]
                  print_message(inspect, f"Address Retreading detected for Worker ID {hex(id_i)}! To avoid address repetition please use a safe value for iters: current={iters} safe={cmd_iters_max_no_retread}")
            if addr_alu is None:
               raise ValueError(f'Unrecognized address preset "{addr}"')

            # Instruction resume for address pattern
            if resume_i:
               start_addr = None
               addr_alu   = None

            # Create address op object
            axaddr = addr_gen(start=start_addr, lo=lo_addr, hi=hi_addr, align=addr_align, alu=addr_alu)

         # Data pattern
         xdata = None
         # Skip data checking
         if data_i is None:
            xdata = None
         # Raw data value
         elif isinstance(data_i, int):
            xdata = data_i
         # Data pattern preset
         elif isinstance(data_i, str):
            # Data equals address
            if data_i == "addr":
               xdata = data_eq_addr()
            # Data equals AxID
            elif data_i == "id":
               xdata = data_eq_id()
            # Clock pattern (every DQ pin alternates between 1 and 0)
            elif data_i == "clock":
               xdata = data_eq_dq(
                  dq0_start  = 0x55555555, dq0_alu  = [dq_echo()]
               )
            # Long 1s followed by 0
            elif data_i == "pulse0":
               xdata = data_eq_dq(
                  dq0_start  = 0xFEFEFEFE, dq0_alu  = [dq_echo()],
               )
            # Long 0s followed by 1
            elif data_i == "pulse1":
               xdata = data_eq_dq(
                  dq0_start  = 0x00000001, dq0_alu  = [dq_echo()],
               )
            # Walking 0s (within a single byte lane)
            elif data_i == "walking0":
               xdata = data_eq_dq(
                  dq0_start  = 0xFEFEFEFE, dq0_alu  = [dq_rotate()],
                  dq1_start  = 0xFDFDFDFD, dq1_alu  = [dq_rotate()],
                  dq2_start  = 0xFBFBFBFB, dq2_alu  = [dq_rotate()],
                  dq3_start  = 0xF7F7F7F7, dq3_alu  = [dq_rotate()],
                  dq4_start  = 0xEFEFEFEF, dq4_alu  = [dq_rotate()],
                  dq5_start  = 0xDFDFDFDF, dq5_alu  = [dq_rotate()],
                  dq6_start  = 0xBFBFBFBF, dq6_alu  = [dq_rotate()],
                  dq7_start  = 0x7F7F7F7F, dq7_alu  = [dq_rotate()],
               )
            # Walking 1s (within a single byte lane)
            elif data_i == "walking1":
               xdata = data_eq_dq(
                  dq0_start  = 0x01010101, dq0_alu  = [dq_rotate()],
                  dq1_start  = 0x02020202, dq1_alu  = [dq_rotate()],
                  dq2_start  = 0x04040404, dq2_alu  = [dq_rotate()],
                  dq3_start  = 0x08080808, dq3_alu  = [dq_rotate()],
                  dq4_start  = 0x10101010, dq4_alu  = [dq_rotate()],
                  dq5_start  = 0x20202020, dq5_alu  = [dq_rotate()],
                  dq6_start  = 0x40404040, dq6_alu  = [dq_rotate()],
                  dq7_start  = 0x80808080, dq7_alu  = [dq_rotate()]
               )
            # PRBS7
            elif data_i == "prbs7":
               xdata = data_eq_dq(
                  dq0_start  = 0x5A5A5A5A, dq0_alu  = [dq_prbs(7)],
                  dq1_start  = 0xC3C3C3C3, dq1_alu  = [dq_prbs(7)],
                  dq2_start  = 0x2B2B2B2B, dq2_alu  = [dq_prbs(7)],
                  dq3_start  = 0x96969696, dq3_alu  = [dq_prbs(7)],
                  dq4_start  = 0x5A5A5A5A, dq4_alu  = [dq_prbs(7)],
                  dq5_start  = 0xC3C3C3C3, dq5_alu  = [dq_prbs(7)],
                  dq6_start  = 0x2B2B2B2B, dq6_alu  = [dq_prbs(7)],
                  dq7_start  = 0x96969696, dq7_alu  = [dq_prbs(7)],
                  dq8_start  = 0x5A5A5A5A, dq8_alu  = [dq_prbs(7)],
                  dq9_start  = 0xC3C3C3C3, dq9_alu  = [dq_prbs(7)],
                  dq10_start = 0x2B2B2B2B, dq10_alu = [dq_prbs(7)],
                  dq11_start = 0x96969696, dq11_alu = [dq_prbs(7)],
                  dq12_start = 0x5A5A5A5A, dq12_alu = [dq_prbs(7)],
                  dq13_start = 0xC3C3C3C3, dq13_alu = [dq_prbs(7)],
                  dq14_start = 0x2B2B2B2B, dq14_alu = [dq_prbs(7)],
                  dq15_start = 0x96969696, dq15_alu = [dq_prbs(7)]
               )
            # PRBS15
            elif data_i == "prbs15":
               xdata = data_eq_dq(
                  dq0_start  = 0x5A5A5A5A, dq0_alu  = [dq_prbs(15)],
                  dq1_start  = 0xFFFFFFFF, dq1_alu  = [dq_prbs(15)],
                  dq2_start  = 0xAAAAAAAA, dq2_alu  = [dq_prbs(15)],
                  dq3_start  = 0x55555555, dq3_alu  = [dq_prbs(15)],
                  dq4_start  = 0x12347890, dq4_alu  = [dq_prbs(15)],
                  dq5_start  = 0x23458901, dq5_alu  = [dq_prbs(15)],
                  dq6_start  = 0x89071234, dq6_alu  = [dq_prbs(15)],
                  dq7_start  = 0xABCDE123, dq7_alu  = [dq_prbs(15)],
                  dq8_start  = 0xFFFFAAAA, dq8_alu  = [dq_prbs(15)],
                  dq9_start  = 0x90875431, dq9_alu  = [dq_prbs(15)],
                  dq10_start = 0x89898989, dq10_alu = [dq_prbs(15)],
                  dq11_start = 0x90213456, dq11_alu = [dq_prbs(15)],
                  dq12_start = 0x5A678901, dq12_alu = [dq_prbs(15)],
                  dq13_start = 0xA5A5A5A5, dq13_alu = [dq_prbs(15)],
                  dq14_start = 0xBCDEFFFF, dq14_alu = [dq_prbs(15)],
                  dq15_start = 0xAAAA5555, dq15_alu = [dq_prbs(15)]
               )
            # PRBS23
            elif data_i == "prbs23":
               xdata = data_eq_dq(
                  dq0_start  = 0x5A5A5A5A, dq0_alu  = [dq_prbs(23)],
                  dq1_start  = 0xFFFFFFFF, dq1_alu  = [dq_prbs(23)],
                  dq2_start  = 0xAAAAAAAA, dq2_alu  = [dq_prbs(23)],
                  dq3_start  = 0x55555555, dq3_alu  = [dq_prbs(23)],
                  dq4_start  = 0x12347890, dq4_alu  = [dq_prbs(23)],
                  dq5_start  = 0x23458901, dq5_alu  = [dq_prbs(23)],
                  dq6_start  = 0x89071234, dq6_alu  = [dq_prbs(23)],
                  dq7_start  = 0xABCDE123, dq7_alu  = [dq_prbs(23)],
                  dq8_start  = 0xFFFFAAAA, dq8_alu  = [dq_prbs(23)],
                  dq9_start  = 0x90875431, dq9_alu  = [dq_prbs(23)],
                  dq10_start = 0x89898989, dq10_alu = [dq_prbs(23)],
                  dq11_start = 0x90213456, dq11_alu = [dq_prbs(23)],
                  dq12_start = 0x5A678901, dq12_alu = [dq_prbs(23)],
                  dq13_start = 0xA5A5A5A5, dq13_alu = [dq_prbs(23)],
                  dq14_start = 0xBCDEFFFF, dq14_alu = [dq_prbs(23)],
                  dq15_start = 0xAAAA5555, dq15_alu = [dq_prbs(23)]
               )
            # PRBS31
            elif data_i == "prbs31":
               xdata = data_eq_dq(
                  dq0_start  = 0x5A5A5A5A, dq0_alu  = [dq_prbs(31)],
                  dq1_start  = 0xFFFFFFFF, dq1_alu  = [dq_prbs(31)],
                  dq2_start  = 0xAAAAAAAA, dq2_alu  = [dq_prbs(31)],
                  dq3_start  = 0x55555555, dq3_alu  = [dq_prbs(31)],
                  dq4_start  = 0x12347890, dq4_alu  = [dq_prbs(31)],
                  dq5_start  = 0x23458901, dq5_alu  = [dq_prbs(31)],
                  dq6_start  = 0x89071234, dq6_alu  = [dq_prbs(31)],
                  dq7_start  = 0xABCDE123, dq7_alu  = [dq_prbs(31)],
                  dq8_start  = 0xFFFFAAAA, dq8_alu  = [dq_prbs(31)],
                  dq9_start  = 0x90875431, dq9_alu  = [dq_prbs(31)],
                  dq10_start = 0x89898989, dq10_alu = [dq_prbs(31)],
                  dq11_start = 0x90213456, dq11_alu = [dq_prbs(31)],
                  dq12_start = 0x5A678901, dq12_alu = [dq_prbs(31)],
                  dq13_start = 0xA5A5A5A5, dq13_alu = [dq_prbs(31)],
                  dq14_start = 0xBCDEFFFF, dq14_alu = [dq_prbs(31)],
                  dq15_start = 0xAAAA5555, dq15_alu = [dq_prbs(31)]
               )

            # Instruction resume for data pattern.
            # This only works for DQ ALU, not the other data modes. Besides, the
            # other data modes don't need this feature.
            if resume_i and data_i != "addr" and data_i != "id":
               xdata = data_eq_dq(
                  dq0_start  = None, dq0_alu  = None
               )

         # Strb pattern
         xstrb = None
         # Raw strb value
         if isinstance(strb_i, int):
            xstrb = strb_i
         # Strb pattern preset
         elif isinstance(strb_i, str):
            # PRBS7
            if strb_i == "prbs7":
               xstrb = strb_eq_dm(
                  dm0_start  = 0x5A5A5A5A, dm0_alu  = [dm_prbs(7)],
                  dm1_start  = 0xC3C3C3C3, dm1_alu  = [dm_prbs(7)],
                  dm2_start  = 0x2B2B2B2B, dm2_alu  = [dm_prbs(7)],
                  dm3_start  = 0x96969696, dm3_alu  = [dm_prbs(7)],
                  dm4_start  = 0x5A5A5A5A, dm4_alu  = [dm_prbs(7)],
                  dm5_start  = 0xC3C3C3C3, dm5_alu  = [dm_prbs(7)],
                  dm6_start  = 0x2B2B2B2B, dm6_alu  = [dm_prbs(7)],
                  dm7_start  = 0x96969696, dm7_alu  = [dm_prbs(7)]
               )
            # PRBS15
            elif strb_i == "prbs15":
               xstrb = strb_eq_dm(
                  dm0_start  = 0x5A5A5A5A, dm0_alu  = [dm_prbs(15)],
                  dm1_start  = 0xFFFFFFFF, dm1_alu  = [dm_prbs(15)],
                  dm2_start  = 0xAAAAAAAA, dm2_alu  = [dm_prbs(15)],
                  dm3_start  = 0x55555555, dm3_alu  = [dm_prbs(15)],
                  dm4_start  = 0x12347890, dm4_alu  = [dm_prbs(15)],
                  dm5_start  = 0x23458901, dm5_alu  = [dm_prbs(15)],
                  dm6_start  = 0x89071234, dm6_alu  = [dm_prbs(15)],
                  dm7_start  = 0xABCDE123, dm7_alu  = [dm_prbs(15)]
               )
            # PRBS23
            elif strb_i == "prbs23":
               xstrb = strb_eq_dm(
                  dm0_start  = 0x5A5A5A5A, dm0_alu  = [dm_prbs(23)],
                  dm1_start  = 0xFFFFFFFF, dm1_alu  = [dm_prbs(23)],
                  dm2_start  = 0xAAAAAAAA, dm2_alu  = [dm_prbs(23)],
                  dm3_start  = 0x55555555, dm3_alu  = [dm_prbs(23)],
                  dm4_start  = 0x12347890, dm4_alu  = [dm_prbs(23)],
                  dm5_start  = 0x23458901, dm5_alu  = [dm_prbs(23)],
                  dm6_start  = 0x89071234, dm6_alu  = [dm_prbs(23)],
                  dm7_start  = 0xABCDE123, dm7_alu  = [dm_prbs(23)]
               )
            # PRBS31
            elif strb_i == "prbs31":
               xstrb = strb_eq_dm(
                  dm0_start  = 0x5A5A5A5A, dm0_alu  = [dm_prbs(31)],
                  dm1_start  = 0xFFFFFFFF, dm1_alu  = [dm_prbs(31)],
                  dm2_start  = 0xAAAAAAAA, dm2_alu  = [dm_prbs(31)],
                  dm3_start  = 0x55555555, dm3_alu  = [dm_prbs(31)],
                  dm4_start  = 0x12347890, dm4_alu  = [dm_prbs(31)],
                  dm5_start  = 0x23458901, dm5_alu  = [dm_prbs(31)],
                  dm6_start  = 0x89071234, dm6_alu  = [dm_prbs(31)],
                  dm7_start  = 0xABCDE123, dm7_alu  = [dm_prbs(31)]
               )
            # Byte-enable test mode. Continually generate random STRB followed by its inverse.
            elif strb_i == "byteen_test":
               xstrb = strb_eq_dm(
                  dm0_start  = 0x5A5A5A5A, dm0_alu  = [dm_invert(), dm_prbs(31)],
                  dm1_start  = 0xFFFFFFFF, dm1_alu  = [dm_invert(), dm_prbs(31)],
                  dm2_start  = 0xAAAAAAAA, dm2_alu  = [dm_invert(), dm_prbs(31)],
                  dm3_start  = 0x55555555, dm3_alu  = [dm_invert(), dm_prbs(31)],
                  dm4_start  = 0x12347890, dm4_alu  = [dm_invert(), dm_prbs(31)],
                  dm5_start  = 0x23458901, dm5_alu  = [dm_invert(), dm_prbs(31)],
                  dm6_start  = 0x89071234, dm6_alu  = [dm_invert(), dm_prbs(31)],
                  dm7_start  = 0xABCDE123, dm7_alu  = [dm_invert(), dm_prbs(31)]
               )

            # Instruction resume for strb pattern.
            if resume_i:
               xstrb = strb_eq_dm(
                  dm0_start  = None, dm0_alu  = None
               )

         # Create worker
         if cmd == 'wr':
            worker = wr_worker(
               awid=axid, awlen=axlen, awsize=axsize, awburst=axburst, awlock=axlock, awcache=axcache, awprot=axprot, awqos=axqos, awregion=axregion, awuser=axuser, awaddr=axaddr,
               wdata=xdata, wstrb=xstrb,
               bresp=xresp
            )
         elif cmd == 'rd':
            worker = rd_worker(
               arid=axid, arlen=axlen, arsize=axsize, arburst=axburst, arlock=axlock, arcache=axcache, arprot=axprot, arqos=axqos, arregion=axregion, aruser=axuser, araddr=axaddr,
               rdata=xdata, rstrb=xstrb, rresp=xresp
            )

         workers.append(worker)

      # Convert bandwidth percentage to duty-cycle tuple
      def percent_to_fraction(p, limit_denominator=8):
         # Max duty-cycle period = 8
         f = Fraction.from_float(p).limit_denominator(limit_denominator)
         return (f.numerator, f.denominator)

      if isinstance(addr_bw, float): addr_bw = percent_to_fraction(addr_bw)
      if isinstance(data_bw, float): data_bw = percent_to_fraction(data_bw)
      if isinstance(resp_bw, float): resp_bw = percent_to_fraction(resp_bw, 64)

      # Generate command
      if cmd == 'wr':
         c = write(iters=iters, awvalid=addr_bw, wvalid=data_bw, bready=resp_bw, awid=id, workers=workers)
      elif cmd == 'rd':
         c = read(iters=iters, arvalid=addr_bw, rready=resp_bw, arid=id, workers=workers)

      # Return the command(s)
      return [c]

   def clog2(self, x):
      '''
      Returns ceil(log(x) / log(2))
      '''
      return math.ceil(math.log(x) / math.log(2))


   def roundup(self, x, y):
      '''
      Returns x rounded up to the next multiple of y
      '''
      return ((x + y-1) // y) * y


   def rounddown(self, x, y):
      '''
      Returns x rounded down to the previous multiple of y
      '''
      return (x // y) * y


class CsrAxi4lDriverPrograms:
   def __init__(self, config, hydra_ip_name, driver_index):
      ip_config   = config['system']['ips'][hydra_ip_name]
      params      = ip_config['parameters']

      num_drivers = int(params['NUM_DRIVERS']['value'])

      driver_type = params[f'DRIVER_{driver_index}_TYPE_ENUM']['value']
      if driver_type != 'DRIVER_TYPE_CSR_AXI4L':
         raise Exception(f'Driver {driver_index} is of type {driver_type}, this class only supports DRIVER_TYPE_CSR_AXI4L')

      # Some parameters related to AXI port widths
      self.max_awaddr = 2**int(params[f'DRIVER_{driver_index}_CSR_AXI4L_AWADDR_WIDTH']['value']) - 1
      self.max_araddr = 2**int(params[f'DRIVER_{driver_index}_CSR_AXI4L_AWADDR_WIDTH']['value']) - 1

      # Traffic patterns accessing HBM registers

      axilite_base_addr = 0x400000000

      #  Dictionary of per-channel hbm registers
      regmap_channel_hbm = {
          "INIT_CTRL"               : {"addr" : 0x00010, "data" : 0x00000000},
          "INIT_STS"                : {"addr" : 0x00014, "data" : 0x00000102},
          "THERM_STS"               : {"addr" : 0x00100, "data" : 0x30000000},
          "IRQ_CTL_PC0"             : {"addr" : 0x00434, "data" : 0x00000200},
          "IRQ_CTL_PC1"             : {"addr" : 0x00734, "data" : 0x00000200},
          "IRQ_EVNTGRP_CTRLSTS_PC0" : {"addr" : 0x00460, "data" : 0x1c7c0000},
          "IRQ_EVNTGRP_CTRLSTS_PC1" : {"addr" : 0x00760, "data" : 0x1c7c0000},
          "AXI4_ERR_STS_PC0"        : {"addr" : 0x00500, "data" : 0x00000000},
          "AXI4_ERR_STS_PC1"        : {"addr" : 0x00800, "data" : 0x00000000},
          "MEM_ERR_STS_PC0"         : {"addr" : 0x00504, "data" : 0x00000000},
          "MEM_ERR_STS_PC1"         : {"addr" : 0x00804, "data" : 0x00000000},
          "ERR_MSK_PC0"             : {"addr" : 0x00520, "data" : 0x00000606},
          "ERR_MSK_PC1"             : {"addr" : 0x00820, "data" : 0x00000606},
          "AWCMDLOG0_PC0"           : {"addr" : 0x00540, "data" : 0x00000000},
          "AWCMDLOG0_PC1"           : {"addr" : 0x00840, "data" : 0x00000000},
          "AWCMDLOG1_PC0"           : {"addr" : 0x00544, "data" : 0x00000000},
          "AWCMDLOG1_PC1"           : {"addr" : 0x00844, "data" : 0x00000000},
          "AWCMDLOG2_PC0"           : {"addr" : 0x00548, "data" : 0x00000000},
          "AWCMDLOG2_PC1"           : {"addr" : 0x00848, "data" : 0x00000000},
          "AWCMDLOG3_PC0"           : {"addr" : 0x0054c, "data" : 0x00000000},
          "AWCMDLOG3_PC1"           : {"addr" : 0x0084c, "data" : 0x00000000},
          "ARCMDLOG0_PC0"           : {"addr" : 0x00550, "data" : 0x00000000},
          "ARCMDLOG0_PC1"           : {"addr" : 0x00850, "data" : 0x00000000},
          "ARCMDLOG1_PC0"           : {"addr" : 0x00554, "data" : 0x00000000},
          "ARCMDLOG1_PC1"           : {"addr" : 0x00854, "data" : 0x00000000},
          "ARCMDLOG2_PC0"           : {"addr" : 0x00558, "data" : 0x00000000},
          "ARCMDLOG2_PC1"           : {"addr" : 0x00858, "data" : 0x00000000},
          "ARCMDLOG3_PC0"           : {"addr" : 0x0055c, "data" : 0x00000000},
          "ARCMDLOG3_PC1"           : {"addr" : 0x0085c, "data" : 0x00000000},
          "ECC_ERR_CNTRS_PC0"       : {"addr" : 0x00560, "data" : 0x00000000},
          "ECC_ERR_CNTRS_PC1"       : {"addr" : 0x00860, "data" : 0x00000000},
      }

      #  Dictionary of per-channel hbm registers
      regmap_hbm = {}
      for k, v in regmap_channel_hbm.items():
          name = k
          addr = v["addr"]
          data = v["data"]
          for ch in [0, 1]:
              name_ch = "CH{:1d}_".format(ch) + name
              ch_addr = 0 if ch == 0 else 1 << 16
              regmap_hbm[name_ch] = {"addr" : ch_addr + addr, "data" : data}

      #  Construct traffic pattern
      addr = axilite_base_addr + regmap_hbm["CH0_ERR_MSK_PC0"]["addr"]
      self.hbm_simulation = [
         #  Read ERR_MSK_PC0, checking against default value
         read_csr(araddr=addr, rdata=0x00606),
         #  Write to ERR_MSK_PC0
         write_csr(awaddr=addr, wdata=0x10606),
         #  Read back written value from ERR_MSK_PC0
         read_csr(araddr=addr, rdata=0x10606),
      ]

      #  Read all CSRs, verifying there are no read errors
      for k, v in regmap_hbm.items():
         addr, data = v["addr"] + axilite_base_addr, v["data"]
         self.hbm_simulation.append(read_csr(araddr=addr))

      #  Write default values to all CSRs, verifying there are no write errors
      for k, v in regmap_hbm.items():
         addr, data = v["addr"] + axilite_base_addr, v["data"]
         self.hbm_simulation.append(write_csr(awaddr=addr, wdata=data))

      # #######
      #  MemSS out-of-the box traffic pattern

      memss_csr_base_addr = 2**(int(params[f'DRIVER_{driver_index}_CSR_AXI4L_AWADDR_WIDTH']['value'])-1)

      self.memss_default = [
         #  Read from DFH first
         read_csr (araddr=axilite_base_addr + 0x0000, rdata=0x00001000),
         read_csr (araddr=axilite_base_addr + 0x0004, rdata=0x30100100),
         read_csr (araddr=axilite_base_addr + 0x0008, rdata=0x043440d8),
         read_csr (araddr=axilite_base_addr + 0x000c, rdata=0x9a11c4dd),
         read_csr (araddr=axilite_base_addr + 0x0010, rdata=0xe9204d00),
         read_csr (araddr=axilite_base_addr + 0x0014, rdata=0x0f43b52e),
         read_csr (araddr=axilite_base_addr + 0x0018, rdata=memss_csr_base_addr),
         read_csr (araddr=axilite_base_addr + 0x001c, rdata=0x00000000),
         read_csr (araddr=axilite_base_addr + 0x0024, rdata=memss_csr_base_addr),
         #  Read and Write to MSS CSR
         read_csr (araddr=axilite_base_addr + memss_csr_base_addr + 0x0000, rdata=0x00010000),
         #  scratchpad
         write_csr(awaddr=axilite_base_addr + memss_csr_base_addr + 0x0020, wdata=0x5a5a5a5a),
         read_csr (araddr=axilite_base_addr + memss_csr_base_addr + 0x0020, rdata=0x5a5a5a5a),
      ]


class CamAxi4DriverPrograms:
   def __init__(self, config, hydra_ip_name, driver_index):
      ip_config   = config['system']['ips'][hydra_ip_name]
      params      = ip_config['parameters']

      num_drivers = int(params['NUM_DRIVERS']['value'])

      driver_type = params[f'DRIVER_{driver_index}_TYPE_ENUM']['value']
      if driver_type != 'DRIVER_TYPE_CAM_AXI4ST':
         raise Exception(f'Driver {driver_index} is of type {driver_type}, this class only supports DRIVER_TYPE_CAM_AXI4ST')

