# (C) 2001-2024 Intel Corporation. All rights reserved.
# Your use of Intel Corporation's design tools, logic functions and other 
# software and tools, and its AMPP partner logic functions, and any output 
# files from any of the foregoing (including device programming or simulation 
# files), and any associated documentation or information are expressly subject 
# to the terms and conditions of the Intel Program License Subscription 
# Agreement, Intel FPGA IP License Agreement, or other applicable 
# license agreement, including, without limitation, that your use is for the 
# sole purpose of programming logic devices manufactured by Intel and sold by 
# Intel or its authorized distributors.  Please refer to the applicable 
# agreement for further details.


######################################################################
# Test Engine IP – System Console library
######################################################################
#
# FILE DESCRIPTION
# ----------------
# This file contains:
# - System Console functions to interact with Test Engine IP on hardware over a JTAG connection
#
# DISCLAIMERS
# ----------------
# Test Engine IP is provided as a beta version. Its functionality and System Console functions are subject to
# change in future releases. Any scripts using this library will need updating if
# function names and behaviors change in future releases.
# This library also is limited to designs with 64 or less interfaces being driven by Test Engine
#
# COMMAND LINE USAGE
# ----------------
# Run system-console --script=testengine_library.tcl --sof=ed_synth.sof --update=1 --n-loops=4
#
# LIBRARY USAGE
# ----------------
# 1) Ensure your design is loaded in system console
# 2) Source this script in system console
#
# SETUP CHECKLIST
# ----------------
# 1) Ensure the Test Engine IP Configuration interface is set to Remote access via JTAG (default in Example designs)
# 2) Ensure that the design is programmed on board
#
# FUNCTIONS
# ----------------
# testengine_num_drivers {} : Return list of AXI4 drivers in the Test Engine. If none found return -1
# testengine_driver_list {} : Return list of AXI4 drivers found in the Traffic Program Directory. If none found return -1
# testengine_reset {} : Reset all Test Engine AXI4 drivers.
# testengine_reprogram {dir} : Program the .hex files in “dir” directory (default dir=”bin”) into Test Engine’s instruction memory.
# testengine_run {} : Run traffic for all Test Engine AXI4 drivers.
# testengine_status {} : Print the status of all Test Engine AXI4 drivers.
# testengine_get_errors {} : Return a list of all Test Engine AXI4 drivers that have raised an error.
# testengine_get_done {} : Return a list of all Test Engine AXI4 drivers that have finished running traffic.
#
# EXAMPLE
# ----------------
# Recommended sequence of function calls to run (and re-run) traffic:
# testengine_reset  OR  testengine_reprogram   #reset while preserving the traffic program, OR reset and program a new one
# #(optional) perform any CSR updates here
# testengine_run                               #run traffic
# testengine_status                            #view status
#



set ::g_update 1
set ::g_num_loops 0
set ::g_number_driver 50
set ::g_sof ""

if {[info exists argc] && [info exists argv]} {
    set n 0
    while {$argc > $n} {
       switch -glob -- [lindex $argv $n] {
          --sof=*               { set ::g_sof                   [string range [lindex $argv $n] [string length "--sof="] end]  }
          --update=*            { set ::g_update                [string range [lindex $argv $n] [string length "--update="] end] }
          --n-loops=*           { set ::g_num_loops             [string range [lindex $argv $n] [string length "--n-loops="] end] }
          --                    { break }
          default               { return -code error "Bad option \"[lindex $argv $n]\". Must be one of --sof=* --update=* --n-loops=*" }
       }
       incr n
    }

    if {$::g_sof == ""} {
        puts "No sof chosen, please load sof manually or use --sof=* option to choose sof. Ignore if sourcing in system-console"
    } else {
        puts "Loading sof: $::g_sof"
        if { [catch {design_load $::g_sof} message] } {
            puts "$message"
        }
    }
}

global bin_found
set bin_found 0
global traffic_program_dir
set traffic_program_dir "bin"
global claim_testengine_jamb_path

set path [lindex [get_service_paths master] [lsearch -glob [get_service_paths master] *remote_access_jamb*]]
if { $path == ""} {
   puts "Error : Test Engine remote JTAG interface not found"
   puts "Ensure .sof is loaded and JTAG access is functional before loading Test Engine System console library"
   set claim_testengine_jamb_path ""
   return
} else {
    puts "Test Engine remote JTAG interface found"
    set claim_testengine_jamb_path [claim_service master $path {} {}]
}

# TEST ENGINE FUNCTIONS

# testengine_num_drivers {} : Returns list of drivers in the Test Engine.
proc testengine_num_drivers {} {
    global claim_testengine_jamb_path
    if {$claim_testengine_jamb_path == ""} {
        puts "Error : Test Engine remote JTAG interface not found"
        puts "Check if design is loaded and source the Test Engine IP System Console library again"
        return -1
    }
    set num_drivers 0
    set driver_list [list]
    for {set i 0x0} {$i < 32} {incr i} {
        set scratchpad [expr { ($i + 1) * 0x1000000 + 0x0000070} ]
        set ovlp_chk [master_read_32 $claim_testengine_jamb_path [format 0x%x $scratchpad] 1]
        if { $ovlp_chk == 0x0F10BACC } {
            break
        }

        master_write_32 $claim_testengine_jamb_path $scratchpad 0x1234ABCD
        set read0 [master_read_32 $claim_testengine_jamb_path $scratchpad 1]
        master_write_32 $claim_testengine_jamb_path $scratchpad 0x0F10BACC
        set read1 [master_read_32 $claim_testengine_jamb_path $scratchpad 1]

        if { $read0 == 0x1234ABCD && $read1 == 0x0F10BACC } {
            lappend driver_list $num_drivers
            incr num_drivers
        } else {
            break
        }
    }

    for {set i 0x0} {$i < $num_drivers} {incr i} {
       set scratchpad [expr { ($i + 1) * 0x1000000 + 0x0000070} ]
       set scratchpad [format 0x%x $scratchpad]
       master_write_32 $claim_testengine_jamb_path $scratchpad 0x0
    }

    if { $num_drivers == 0 } {
       puts "Error : Failed to discover Test Engine"
       return -1
    }
    return $driver_list
}

proc testengine_driver_list {} {
    global traffic_program_dir
    global claim_testengine_jamb_path

    if {$claim_testengine_jamb_path == ""} {
        puts "Error : Test Engine remote JTAG interface not found"
        puts "Check if design is loaded and source the Test Engine IP System Console library again"
        return -1
    }
    if { [catch { set drivers [glob -directory $traffic_program_dir -- "*"] } errorstring] } {
       puts "Traffic program directory $traffic_program_dir was not found : $errorstring"
       return -1
    }
    set drivers [lsort -dictionary $drivers]
    set driver_list [list]

    foreach driver $drivers {
        regexp {driver(?:(\w+)_mem_axi4)} $driver -> c1
        if {[info exists c1]} {
            lappend driver_list [lindex $c1]
        }
    }
    if {[llength $driver_list] == 0} {
        return -1
    }
    return $driver_list
}

# testengine_global_run {} : Rerun Test Engine Traffic across all drivers through the global CSR path.
proc testengine_run {} {
    global claim_testengine_jamb_path
    global bin_found

    if {$bin_found} {
        set num_drivers [testengine_driver_list]
    } else {
        set num_drivers [testengine_num_drivers]
    }
    if { $num_drivers == -1 } {
        puts "Unable to find number of drivers in the system, try another traffic program"
        return
    }
    puts "Running traffic on all drivers ..."
    set driver_run_bitmask_0_addr 0x0080
    set driver_run_bitmask_1_addr 0x0084
    set driver_run_bitmask_0 [master_read_32 $claim_testengine_jamb_path $driver_run_bitmask_0_addr 1]
    set driver_run_bitmask_1 [master_read_32 $claim_testengine_jamb_path $driver_run_bitmask_1_addr 1]

    set driver_run_bitmask [expr {($driver_run_bitmask_1 << 32) | $driver_run_bitmask_0}]

    set bitmask 0x0
    foreach driver $num_drivers {
        set bit_pos [expr 0x1 << $driver]
        set bitmask [expr $bitmask | $bit_pos]
    }

    set driver_run_bitmask [expr $bitmask | $driver_run_bitmask]
    master_write_32 $claim_testengine_jamb_path $driver_run_bitmask_0_addr [expr $driver_run_bitmask & 0xFFFFFFFF]
    master_write_32 $claim_testengine_jamb_path $driver_run_bitmask_1_addr [expr ($driver_run_bitmask & 0xFFFFFFFF00000000) >> 32]

    set testengine_ctrl_status_0_addr 0x0078
    set testengine_ctrl_status_0 [master_read_32 $claim_testengine_jamb_path $testengine_ctrl_status_0_addr 1]

    if {$testengine_ctrl_status_0 & 0x4} {
        puts "Traffic has already been run."
        puts "Please call testengine_status to view results, and call testengine_reset before invoking testengine_run."
        return
    }

    set testengine_ctrl_status_0 [expr {$testengine_ctrl_status_0 | 0x4}]
    master_write_32 $claim_testengine_jamb_path $testengine_ctrl_status_0_addr $testengine_ctrl_status_0

    puts "Traffic is now running for all drivers"
}


proc testengine_reset {} {
    global claim_testengine_jamb_path
    global bin_found

    if {$bin_found} {
        set num_drivers [testengine_driver_list]
    } else {
        set num_drivers [testengine_num_drivers]
    }
    if { $num_drivers == -1 } {
        puts "Unable to find number of drivers in the system, try another traffic program"
        return
    }

    puts "Resetting all drivers ..."
    set testengine_ctrl_status_0_addr 0x0078
    set testengine_ctrl_status_0 [master_read_32 $claim_testengine_jamb_path $testengine_ctrl_status_0_addr 1]

    set testengine_ctrl_status_0 [expr {$testengine_ctrl_status_0 & 0xFFFFFFFB}]
    master_write_32 $claim_testengine_jamb_path $testengine_ctrl_status_0_addr $testengine_ctrl_status_0

    foreach driver $num_drivers {
        set driver_ctrl_stat_addr [expr { ($driver + 1) * 0x1000000 + 0x0000068} ]
        master_write_32 $claim_testengine_jamb_path $driver_ctrl_stat_addr  0x1
    }

    foreach driver $num_drivers {
        set driver_ctrl_stat_addr [expr { ($driver + 1) * 0x1000000 + 0x0000068} ]
        master_write_32 $claim_testengine_jamb_path $driver_ctrl_stat_addr  0x0
    }

    puts "Reset complete"
}

# testengine_get_errors {} : Return a list of all Test Engine AXI4 drivers that have raised an error.
proc testengine_get_errors {} {

    global claim_testengine_jamb_path
    global bin_found

    if {$bin_found} {
        set num_drivers [testengine_driver_list]
    } else {
        set num_drivers [testengine_num_drivers]
    }

    if { $num_drivers == -1 } {
        puts "Error: No AXI4 drivers found in Test Engine"
        return [list]
    }

    set testengine_error_0  [master_read_32 $claim_testengine_jamb_path 0x00000C0 1]
    set testengine_error_1  [master_read_32 $claim_testengine_jamb_path 0x00000C4 1]
    set testengine_error    [expr {($testengine_error_1  << 32) | $testengine_error_0}]

    set failed_drivers {}
    foreach driver $num_drivers {
        set bit_pos [expr 0x1 << $driver]

        if { $testengine_error & $bit_pos } {
            lappend failed_drivers $driver
        }
    }

    return $failed_drivers
}

# testengine_get_done {} : Return a list of all Test Engine AXI4 drivers that have finished running traffic.
proc testengine_get_done {} {

    global claim_testengine_jamb_path
    global bin_found

    if {$bin_found} {
        set num_drivers [testengine_driver_list]
    } else {
        set num_drivers [testengine_num_drivers]
    }

    if { $num_drivers == -1 } {
        puts "Error: No AXI4 drivers found in Test Engine"
        return [list]
    }

    set testengine_done_0 [master_read_32 $claim_testengine_jamb_path 0x00000A0 1]
    set testengine_done_1 [master_read_32 $claim_testengine_jamb_path 0x00000A4 1]
    set testengine_done   [expr {($testengine_done_1  << 32) | $testengine_done_0}]

    set done_drivers {}
    foreach driver $num_drivers {
        set bit_pos [expr 0x1 << $driver]

        if { $testengine_done & $bit_pos } {
            lappend done_drivers $driver
        }
    }

    return $done_drivers
}

# testengine_status {} : Read the status of all hydra drivers through the global CSR and print a summary, and returns a driver list.
proc testengine_status {} {

    global claim_testengine_jamb_path
    global bin_found

    if {$bin_found} {
        set num_drivers [testengine_driver_list]
    } else {
        set num_drivers [testengine_num_drivers]
    }
    if { $num_drivers == -1 } {
        puts "Error: No AXI4 drivers found in Test Engine"
        return
    }


    set testengine_done_0 [master_read_32 $claim_testengine_jamb_path 0x00000A0 1]
    set testengine_done_1 [master_read_32 $claim_testengine_jamb_path 0x00000A4 1]
    set testengine_done   [expr {($testengine_done_1  << 32) | $testengine_done_0}]
    set testengine_error_0  [master_read_32 $claim_testengine_jamb_path 0x00000C0 1]
    set testengine_error_1  [master_read_32 $claim_testengine_jamb_path 0x00000C4 1]
    set testengine_error    [expr {($testengine_error_1  << 32) | $testengine_error_0}]
    set drivers_done   0
    set drivers_passed 0
    set unfin_drivers  {}
    set failed_drivers {}

    set current_driver_done 0

    foreach driver $num_drivers {
        set bit_pos [expr 0x1 << $driver]

        if { $testengine_done & $bit_pos } {
            incr drivers_done
            set current_driver_done 1
        } else {
            set current_driver_done 0
            lappend unfin_drivers $driver
        }

        if { $testengine_error & $bit_pos } {
            lappend failed_drivers $driver
        } elseif {$current_driver_done} {
            incr drivers_passed
        }
    }
    puts "Driver done: $drivers_done / [llength $num_drivers]"
    if {[llength $unfin_drivers] != 0 } {
        set unfin_str [lindex $unfin_drivers 0]
       for {set i 1} {$i < [llength $unfin_drivers]} {incr i} {
          append unfin_str ", " [lindex $unfin_drivers $i]
       }
       puts "Unfinished Drivers: ($unfin_str)"
    }
    puts "Driver pass: $drivers_passed / [llength $num_drivers]"
    if { [llength $failed_drivers] != 0 } {
       set fail_str [lindex $failed_drivers 0]
       for {set i 1} {$i < [llength $failed_drivers]} {incr i} {
          append fail_str ", " [lindex $failed_drivers $i]
       }
       puts "Failed Drivers: ($fail_str)"
    }

    if {$drivers_done == 0 && [llength $failed_drivers] == 0} {
        set testengine_ctrl_status_0_addr 0x0078
        set testengine_ctrl_status_0 [master_read_32 $claim_testengine_jamb_path $testengine_ctrl_status_0_addr 1]
        if {$testengine_ctrl_status_0 & 0x4} {
            return
        } else {
            puts "No Traffic was run"
            puts "Please call testengine_run after reset to see traffic results"
            return
        }
    }
}

# testengine_reprogram {dir} : Program the .hex files in “dir” directory (default dir=”bin”) into Test Engine’s instruction memory.
proc testengine_reprogram {{dir "bin"} args} {
    global claim_testengine_jamb_path
    global bin_found
    global traffic_program_dir

    set old_dir $traffic_program_dir
    set old_bin_found $bin_found
    set bin_found 1
    set traffic_program_dir $dir

    set driver_list [testengine_driver_list]

    if { $driver_list == -1 } {
        puts "Unable to find number of drivers in the system, try another traffic program"
        set bin_found $old_bin_found
        set traffic_program_dir $old_dir
        return
    } else {
        puts "Traffic program directory updated to $dir"
    }

    foreach driver $driver_list {
        set driver_ctrl_stat_addr [expr { ($driver + 1) * 0x1000000 + 0x0000068} ]
        master_write_32 $claim_testengine_jamb_path $driver_ctrl_stat_addr  0x0
    }

    set testengine_ctrl_status_0_addr 0x0078
    set testengine_ctrl_status_0 [master_read_32 $claim_testengine_jamb_path $testengine_ctrl_status_0_addr 1]

    set testengine_ctrl_status_0 [expr {$testengine_ctrl_status_0 & 0xFFFFFFFB}]
    master_write_32 $claim_testengine_jamb_path $testengine_ctrl_status_0_addr $testengine_ctrl_status_0

    foreach driver $driver_list {
        if { [catch { set instr_rams_commands_txt_files [glob -directory ${traffic_program_dir}/driver${driver}_mem_axi4/csr -- "*.txt"] } errorstring] } {
            puts "Traffic program files for driver $driver were not found : $errorstring"
        } else {
            set empty [catch {file size $instr_rams_commands_txt_files} size]
            if {$empty == 0 && $size > 0} {
                puts "Traffic program file $instr_rams_commands_txt_files for driver ${traffic_program_dir}/driver${driver}_mem_axi4 was empty"
            } else {
                puts "Updating instruction memory for driver $driver... this can take ~1 minute"

                set base_addr [expr { ($driver + 1) * 0x1000000}]

                foreach instr_rams_commands_txt_file $instr_rams_commands_txt_files {
                    set file [open $instr_rams_commands_txt_file r]
                    while {[gets $file line]>=0} {
                        set addr [lindex $line 0]
                        set data_32 [lindex $line 1]
                        set addr [expr {$addr | $base_addr}]
                        master_write_32 $claim_testengine_jamb_path $addr [list $data_32]
                    }
                    close $file
                }
            }

        }
        incr driver_idx
    }

    testengine_reset

    puts "Instruction memory update complete"
}

puts "Test Engine System Console library loaded"


if { [catch { set driver_list [glob -directory "bin" -- "*"] } errorstring] } {
    set dir_list [glob -type d "*"]
    foreach dir $dir_list {
        set traffic_program_dir $dir
        set te_driver_list [testengine_driver_list]
        if { $te_driver_list != -1 } {
            set bin_found 1
            puts "Traffic program directory $dir was found"
            puts "If this traffic program does not match the design on board run testengine_reprogram <dir> "
            break
        }
    }
    if { $bin_found == 0}  {
        set traffic_program_dir "bin"
        puts "No Traffic program directory was found : $errorstring"
    }
} else {
    set bin_found 1
    puts "Traffic program directory bin was found"
}

# system-console --script=testengine_library.tcl test

if {[info exists argc] && [info exists argv]} {
    set ::updated 0

    for { set loop 0 } { $loop < $::g_num_loops } { incr loop } {

        puts "\n****    Loop $loop    ****"
        testengine_reset

        if { $::g_update &&  ($::updated == "0" ) } {
            testengine_reprogram
            set ::updated 1
        }

        testengine_run
        testengine_status
    }
}

# BETA FUNCTION
# testengine_driver_parse_errors {driver} {dir} :
# Parses errors from the AXI4 driver chosen and prints them to a log file under the chosen directory labelled driver_${driver}_errors.txt if any errors are found.
# Defaults to driver index 0 if no index is provided and the current directory if no directory is chosen.
proc testengine_driver_parse_errors { {driver 0} {log_dir "."} } {
    global claim_testengine_jamb_path
    global bin_found

    if {$bin_found} {
        set num_drivers [testengine_driver_list]
    } else {
        set num_drivers [testengine_num_drivers]
    }

    if {[lsearch -exact $num_drivers $driver] == -1} {
        puts "Error: $driver is not a valid choice for an AXI4 driver index"
        return
    }


    if { [catch {set log [open "${log_dir}/driver_${driver}_errors.txt" w+]} fid] } {
        puts "Error: Invalid directory $log_dir provided for writing\n$fid"
        return
    }

    puts $log "Driver $driver :"
    puts $log ""

    set driver_base_addr [expr { ($driver + 1) * 0x1000000}]

    puts "Reading from driver $driver error log"
    set csrs [master_read_32 $claim_testengine_jamb_path [expr {$driver_base_addr + 0x98}] 51]
    puts "Writing to results to ${log_dir}/driver_${driver}_errors.txt"
    puts $log "Num BID errors = [lindex $csrs 0]"
    puts $log "Num BRESP errors = [lindex $csrs 1]"
    puts $log "Num RID errors = [lindex $csrs 2]"
    puts $log "Num RRESP errors = [lindex $csrs 3]"
    puts $log "Num RDATA errors = [lindex $csrs 4]"
    puts $log "Num RLAST errors = [lindex $csrs 5]"
    set val 0
    for {set i 45} {$i >= 6} {incr i -1} {
        set val "${val}[string range [lindex $csrs $i] 2 end]"
    }
    puts $log "PNF = 0x$val"
    set val 0
    for {set i 49} {$i >= 46} {incr i -1} {
        set val "${val}[string range [lindex $csrs $i] 2 end]"
    }
    puts $log "TER DQ mask = 0x$val"
    puts $log "TER (Transaction Error Count) = [lindex $csrs 50]"
























    close $log

    set errors [lrange $csrs 0 5]
    if {[lsearch -not $errors 0x00000000] == -1} {
        puts "Error logging complete. No errors found in log."
    } else {
        puts "Error logging complete. Errors found in log."
    }
}

